#!/usr/bin/env python
""" compare how the elements behave

We create a folder ground truth that possess the same thing than the other
in a form of a dictionnary containing nparrays and other info.
the other files contains every test and the name is the date of the test

See Also
------------

Link 

\image dev/kalfon/img/datacomparison.pdf
\author: jeremie KALFON
\date Created on Tue Jun 30 21:01:17 2015
\copyright GNU General Public License v2.0
\package CaImAn/comparison
"""
#
#\image html dev/kalfon/img/datacomparison.pdf
#\version   1.0
#
#
#

import copy
import datetime
import logging
import matplotlib.pyplot as pl
import numpy as np
import os
import platform as plt
import scipy
from typing import Dict

# Set up the logger; change this if you like.
# You can log to a file using the filename parameter, or make the output more or less
# verbose by setting level to logging.DEBUG, logging.INFO, logging.WARNING, or logging.ERROR

logging.basicConfig(
    format="%(relativeCreated)12d [%(filename)s:%(funcName)20s():%(lineno)s] [%(process)d] %(message)s",
                                                                                                         # filename="/tmp/caiman.log",
    level=logging.DEBUG)

import caiman as cm
from caiman.paths import caiman_datadir


class Comparison(object):
    """
       Comparison(object): It is used for comparison the modification on the CaImAn program


       You need to compare your modification to the groundtruth and then rename the folder created as "proof" for it to be send in you next push request

       Here it has been made for 3 different functions. for it to compare well you need to set your
       ground truth with the same computer with which you are comparing the files
       class you instanciate to compare the different parts of CaImAn




        Attributes
        ----------
        self : object
    LIST OF WHAT YOU CAN FIND IN THE SAVED FILE
      data
        A_full
        A_patch
        rig_shifts
        C_full
        C_patch
        information
                diffrences
                        params_cnm
                        proc
                        param_movie
                params
                timer
                        rig_shifts
                        cnmf on patch
                        cnmf_full_frame
                time
                processor
                cnmpatch
                platform
                diff
                        cnmfull [same parameters](#732ed9)
                        cnmpatch
                                diff_data
                                        performance
                                                f1_score
                                                accuracy
                                                precision
                                                recall
                                        diffneur
                                        correlations
                                isdifferent
                                diff_timing
                        rig
                                diff_timing
                                isdifferent
                                diff_data
            the user to change it manualy

        Methods
        -------
        __init__()
            Initialize the function be instanciating a comparison object

        save(istruth)
            save the comparison object on a file

        save_with_compare(self, istruth, params, dview, avg_min_size_ratio, n_frames_per_bin, dims_test, dims_gt, Cn, cmap)
            the save functions to use !

        see(self,filename=None)
            to look into a particular saved file

        plotall(self)
            when you want to plot the variations of the different tests that have been compared according to the actual groundtruth

        See Also
        --------

        .. image:: caiman/tests/comparison/data.pdf
        """

    def __init__(self):

        self.comparison:Dict[str, Dict] = {
            'rig_shifts': {},
            'pwrig_shifts': {},
            'cnmf_on_patch': {},
            'cnmf_full_frame': {},
        }

        self.comparison['rig_shifts'] = {
            'ourdata': None,
            'timer': None,
            'sensitivity': 0.001               # the sensitivity USER TO CHOOSE
        }
                                               # apparently pwrig shift are not used any more and the comparison are useless
                                               # self.comparison['pwrig_shifts']={
                                               #                 'ourdata': None,
                                               #                'timer': None,
                                               #               'sensitivity': 0.001
                                               #           }
        self.comparison['cnmf_on_patch'] = {'ourdata': None, 'timer': None, 'sensitivity': 0.01}
        self.comparison['cnmf_full_frame'] = {'ourdata': None, 'timer': None, 'sensitivity': 0.01}

        self.cnmpatch = None
        self.information = None
        self.dims = None

    def save_with_compare(self, istruth=False, params=None, dview=None, Cn=None):
        """save the comparison as well as the images of the precision recall calculations


            depending on if we say this file will be ground truth or not, it wil be saved in either the tests or the ground truth folder
            if saved in test, a comparison to groundtruth will be added to the object 
            this comparison will be on 
                data : a normized difference of the normalized value of the arrays
                time : difference
            in order for this function to work, you must
                have previously given it the cnm objects after initializing them ( on patch and full frame)
                give the values of the time and data 
                have a groundtruth


            Args:
                self:  dictionnary
                   the object of this class tha tcontains every value

                istruth: Boolean
                    if we want it ot be the ground truth

                params:
                    movie parameters

                dview :
                    your dview object

                n_frames_per_bin:
                    you need to know those data before
                    they have been given to the base/rois functions

                dims_test:
                    you need to know those data before
                    they have been given to the base/rois functions

                Cn:
                    your correlation image

                Cmap:
                    a particular colormap for your Cn

            See Also:
                Example of utilisation on Demo Pipeline
\image caiman/tests/comparison/data.pdf

             Raises:
                 ('we now have ground truth\n')

                 ('we were not able to read the file to compare it\n')

                """
        # getting the DATA FOR COMPARISONS
        assert (params != None and self.cnmpatch != None)
        logging.info('we need the parameters in order to save anything\n')
        # actions on the sparse matrix
        cnm = self.cnmpatch.__dict__
        cnmpatch = deletesparse(cnm)

        # initialization
        dims_test = [self.dims[0], self.dims[1]]
        dims_gt = dims_test
        today_dt = datetime.datetime.today()
        today_str = str(today_dt)
        plat = plt.platform()
        plat = str(plat)
        pro = plt.processor()
        pro = str(pro)
        # we store a big file which contains everything (INFORMATION)
        information = {
            'platform': plat,
            'time': today_str,
            'processor': pro,
            'params': params,
            'cnmpatch': cnmpatch,
            'timer': {
                'cnmf_on_patch': self.comparison['cnmf_on_patch']['timer'],
                'cnmf_full_frame': self.comparison['cnmf_full_frame']['timer'],
                'rig_shifts': self.comparison['rig_shifts']['timer']
            }
        }

        file_path = os.path.join(caiman_datadir(), "testdata", "groundtruth.npz")

        # OPENINGS
        # if we want to set this data as truth
        if istruth:
            # we just save it
            if os.path.exists(file_path):
                os.remove(file_path)
            else:
                logging.debug("nothing to remove\n")
            np.savez_compressed(file_path,
                                information=information,
                                A_full=self.comparison['cnmf_full_frame']['ourdata'][0],
                                C_full=self.comparison['cnmf_full_frame']['ourdata'][1],
                                A_patch=self.comparison['cnmf_on_patch']['ourdata'][0],
                                C_patch=self.comparison['cnmf_on_patch']['ourdata'][1],
                                rig_shifts=self.comparison['rig_shifts']['ourdata'])
            logging.info('we now have ground truth\n')
            return

        else:                                                                                               # if not we create a comparison first
            try:
                with np.load(file_path, encoding='latin1', allow_pickle=True) as dt:
                    rig_shifts = dt['rig_shifts'][()]
                    A_patch = dt['A_patch'][()]
                    A_full = dt['A_full'][()]
                    C_full = dt['C_full'][()]
                    C_patch = dt['C_patch'][()]
                    data = dt['information'][()]
                                                                                                            # if we cannot manage to open it or it doesnt exist:
            except (IOError, OSError):
                                                                                                            # we save but we explain why there were a problem
                logging.warning('we were not able to read the file ' + str(file_path) + ' to compare it\n')
                file_path = os.path.join(caiman_datadir(), "testdata", "NC" + dt + ".npz")
                np.savez_compressed(file_path,
                                    information=information,
                                    A_full=self.comparison['cnmf_full_frame']['ourdata'][0],
                                    C_full=self.comparison['cnmf_full_frame']['ourdata'][1],
                                    A_patch=self.comparison['cnmf_on_patch']['ourdata'][0],
                                    C_patch=self.comparison['cnmf_on_patch']['ourdata'][1],
                                    rig_shifts=self.comparison['rig_shifts']['ourdata'])
                return
                                                                                                            # creating the FOLDER to store our data
                                                                                                            # XXX Is this still hooked up to anything?
        i = 0
        dr = os.path.join(caiman_datadir(), "testdata")
        for name in os.listdir(dr):
            i += 1
        istr = str(i)
        if not os.path.exists(dr + istr):
            os.makedirs(dr + istr)
        information.update({'diff': {}})
        information.update({'differences': {'proc': False, 'params_movie': False, 'params_cnm': False}})
                                                                                                            # INFORMATION FOR THE USER
        if data['processor'] != information['processor']:
            logging.info("you don't have the same processor as groundtruth.. the time difference can vary"
                         " because of that\n try recreate your own groundtruth before testing. Compare: " +
                         str(data['processor']) + " to " + str(information['processor']) + "\n")
            information['differences']['proc'] = True
        if data['params'] != information['params']:
            logging.warning("you are not using the same movie parameters... Things can go wrong")
            logging.warning('you must use the same parameters to compare your version of the code with '
                            'the groundtruth one. look for the groundtruth parameters with the see() method\n')
            information['differences']['params_movie'] = True
                                                                                                            # We must cleanup some fields to permit an accurate comparison
        if not normalised_compare_cnmpatches(data['cnmpatch'], cnmpatch):
            if data['cnmpatch'].keys() != cnmpatch.keys():
                logging.error(
                    'DIFFERENCES IN THE FIELDS OF CNMF'
                )                                                                                           # TODO: Now that we have deeply nested data structures, find a module that gives you tight differences.
            diffkeys = [k for k in data['cnmpatch'] if data['cnmpatch'][k] != cnmpatch[k]]
            for k in diffkeys:
                logging.info("{}:{}->{}".format(k, data['cnmpatch'][k], cnmpatch[k]))

            logging.warning('you are not using the same parameters in your cnmf on patches initialization\n')
            information['differences']['params_cnm'] = True

        # for rigid
        # plotting part

        information['diff'].update({
            'rig':
            plotrig(init=rig_shifts,
                    curr=self.comparison['rig_shifts']['ourdata'],
                    timer=self.comparison['rig_shifts']['timer'] - data['timer']['rig_shifts'],
                    sensitivity=self.comparison['rig_shifts']['sensitivity'])
        })
        #try:
        #    pl.gcf().savefig(dr + str(i) + '/' + 'rigidcorrection.pdf')
        #    pl.close()
        #except:
        #    pass

        # for cnmf on patch
        information['diff'].update({
            'cnmpatch':
            cnmf(Cn=Cn,
                 A_gt=A_patch,
                 A_test=self.comparison['cnmf_on_patch']['ourdata'][0],
                 C_gt=C_patch,
                 C_test=self.comparison['cnmf_on_patch']['ourdata'][1],
                 dview=dview,
                 sensitivity=self.comparison['cnmf_on_patch']['sensitivity'],
                 dims_test=dims_test,
                 dims_gt=dims_gt,
                 timer=self.comparison['cnmf_on_patch']['timer'] - data['timer']['cnmf_on_patch'])
        })
        #try:
        #    pl.gcf().savefig(dr + istr + '/' + 'onpatch.pdf')
        #    pl.close()
        #except:
        #    pass

        # CNMF FULL FRAME
        information['diff'].update({
            'cnmfull':
            cnmf(Cn=Cn,
                 A_gt=A_full,
                 A_test=self.comparison['cnmf_full_frame']['ourdata'][0],
                 C_gt=C_full,
                 C_test=self.comparison['cnmf_full_frame']['ourdata'][1],
                 dview=dview,
                 sensitivity=self.comparison['cnmf_full_frame']['sensitivity'],
                 dims_test=dims_test,
                 dims_gt=dims_gt,
                 timer=self.comparison['cnmf_full_frame']['timer'] - data['timer']['cnmf_full_frame'])
        })
        #try:
        #    pl.gcf().savefig(dr + istr + '/' + 'cnmfull.pdf')
        #    pl.close()
        #except:
        #    pass

        # Saving of everything
        target_dir = os.path.join(caiman_datadir(), "testdata", istr)
        if not os.path.exists(target_dir):
            os.makedirs(os.path.join(caiman_datadir(), "testdata", istr)) # TODO Revise to just use the exist_ok flag to os.makedirs
        file_path = os.path.join(target_dir, istr + ".npz")
        np.savez_compressed(file_path,
                            information=information,
                            A_full=self.comparison['cnmf_full_frame']['ourdata'][0],
                            C_full=self.comparison['cnmf_full_frame']['ourdata'][1],
                            A_patch=self.comparison['cnmf_on_patch']['ourdata'][0],
                            C_patch=self.comparison['cnmf_on_patch']['ourdata'][1],
                            rig_shifts=self.comparison['rig_shifts']['ourdata'])

        self.information = information


def see(filename=None):
    """shows you the important data about a certain test file ( just give the number or name)

        if you give nothing it will give you back the groundtruth infos

        Args:
            self:  dictionary
                the object of this class tha tcontains every value
            filename:
                ( just give the number or name)

        See Also:
            @image html caiman/tests/comparison/data.pdf
            """

    if filename == None:
        dr = os.path.join(caiman_datadir(), "testdata", "groundtruth.npz")
    else:
        dr = os.path.join(caiman_datadir(), "testdata", filename, filename + ".npz")
        logging.debug("Loading GT file " + str(dr))
    with np.load(dr) as dt:
        print('Info :\n')
        see_it(dt)


def see_it(data=None):
    for key in data:

        val = data[key]
        if isinstance(val, dict):
            print('\n')
            print(key)
            print('\n')
            see_it(val)
        else:
            if not isinstance(val, scipy.sparse.coo.coo_matrix) \
                    and not isinstance(val, scipy.sparse.csc.csc_matrix) and not isinstance(val, list):

                print(key)
                print(val)


def deletesparse(cnm):
    for keys in cnm:
        val = cnm[keys]
        if isinstance(val, dict):
            val = deletesparse(val)
        if not isinstance(val, scipy.sparse.coo.coo_matrix) and not isinstance(val, np.ndarray) \
                and not isinstance(val, scipy.sparse.csc.csc_matrix) and not keys == 'dview':
            logging.debug("type of val is " + str(type(val)))
            cnm[keys] = val
        else:

            cnm[keys] = None
    return cnm


def cnmf(Cn, A_gt, A_test, C_gt, C_test, dims_gt, dims_test, dview=None, sensitivity=0, timer=0, n_frames_per_bin=10):

    A_test = A_test.toarray()  # coo sparse matrix
    A_gt = A_gt.toarray()

    # proceed to a trhreshold
    A_test_thr = cm.source_extraction.cnmf.spatial.threshold_components(A_test,
                                                                        dims_test,
                                                                        medw=None,
                                                                        thr_method='max',
                                                                        maxthr=0.2,
                                                                        nrgthr=0.99,
                                                                        extract_cc=True,
                                                                        se=None,
                                                                        ss=None,
                                                                        dview=dview)
    A_gt_thr = cm.source_extraction.cnmf.spatial.threshold_components(A_gt,
                                                                      dims_gt,
                                                                      medw=None,
                                                                      thr_method='max',
                                                                      maxthr=0.2,
                                                                      nrgthr=0.99,
                                                                      extract_cc=True,
                                                                      se=None,
                                                                      ss=None,
                                                                      dview=dview)

    # compute C using this A thr
    A_test_thr = A_test_thr.toarray() > 0
    A_gt_thr = A_gt_thr.toarray() > 0
    # we do not compute a threshold on the size of neurons
    C_test_thr = C_test
    C_gt_thr = C_gt
    # we would also like the difference in the number of neurons
    diffneur = A_test_thr.shape[1] - A_gt_thr.shape[1]
    #    print(diffneur+1)
    # computing the values
    C_test_thr = np.array([CC.reshape([-1, n_frames_per_bin]).max(1) for CC in C_test_thr])
    C_gt_thr = np.array([CC.reshape([-1, n_frames_per_bin]).max(1) for CC in C_gt_thr])
    maskgt = A_gt_thr[:, :].reshape([dims_gt[0], dims_gt[1], -1], order='F').transpose([2, 0, 1]) * 1.
    masktest = A_test_thr[:, :].reshape([dims_test[0], dims_test[1], -1], order='F').transpose([2, 0, 1]) * 1.

    idx_tp_gt, idx_tp_comp, idx_fn_gt, idx_fp_comp, performance_off_on =  \
        cm.base.rois.nf_match_neurons_in_binary_masks(masks_gt=maskgt,
                                                      masks_comp=masktest, Cn=Cn, plot_results=False)

    # the pearson's correlation coefficient of the two Calcium activities thresholded
    # comparing Calcium activities of all the components that are defined by

    corrs = np.array(
        [scipy.stats.pearsonr(C_gt_thr[gt, :], C_test_thr[comp, :])[0] for gt, comp in zip(idx_tp_gt, idx_tp_comp)])
    # todo, change this test when I will have found why I have one additionnal neuron

    isdiff = True if ((np.linalg.norm(corrs) < sensitivity) or (performance_off_on['f1_score'] < 0.98)) else False
    info = {
        'isdifferent': int(isdiff),
        'diff_data': {
            'performance': performance_off_on,
            'corelations': corrs.tolist(),
                                                       #performance = dict()
                                                       #performance['recall'] = old_div(TP,(TP+FN))
                                                       #performance['precision'] = old_div(TP,(TP+FP))
                                                       #performance['accuracy'] = old_div((TP+TN),(TP+FP+FN+TN))
                                                       #performance['f1_score'] = 2*TP/(2*TP+FP+FN)
            'diffneur': diffneur
        },
        'diff_timing': timer
    }
    return info


def plotrig(init, curr, timer, sensitivity):

    diff = np.linalg.norm(np.asarray(init) - np.asarray(curr)) / np.linalg.norm(init)
    isdiff = diff > sensitivity
    info = {'isdifferent': int(isdiff), 'diff_data': diff, 'diff_timing': timer}
    curr = np.asarray(curr).transpose([1, 0])
    init = init.transpose([1, 0])
    xc = np.arange(curr.shape[1])
    xi = np.arange(init.shape[1])
    #try:
    #    pl.figure()
    #    pl.subplot(1, 2, 1)
    #    pl.plot(xc, curr[0], 'r', xi, init[0], 'b')
    #    pl.legend(['x shifts curr', 'x shifts init'])
    #    pl.xlabel('frames')
    #    pl.ylabel('pixels')
    #    pl.subplot(1, 2, 2)
    #    pl.plot(xc, curr[1], 'r', xi, init[1], 'b')
    #    pl.legend(['yshifts curr', 'y shifts init'])
    #    pl.xlabel('frames')
    #    pl.ylabel('pixels')
    #except:
    #    logging.warning("not able to plot")
    return info


def normalised_compare_cnmpatches(a, b):
    # This is designed to copy with fields that make it into these objects that need some normalisation before they
    # are rightly comparable. To deal with that we do a deepcopy and then inline-normalise. Add any new needed keys
    # into this code. Right now this is manual, but if we need to do more of this we should turn this into a nice
    # list of fields to ignore, and another list of fields to perform regular transforms on.
    mutable_a = copy.deepcopy(a)
    mutable_b = copy.deepcopy(b)

    if 'params' in mutable_a and 'params' in mutable_b:
        params_a = mutable_a['params']
        params_b = mutable_b['params']
        if hasattr(params_a, 'online') and hasattr(params_b, 'online'):
            if 'path_to_model' in params_a.online and 'path_to_model' in params_b.online:
                _, params_a.online['path_to_model'] = os.path.split(
                    params_a.online['path_to_model'])                  # Remove all but the last part
                _, params_b.online['path_to_model'] = os.path.split(params_b.online['path_to_model'])
                                                                       # print("Normalised A: " + str(params_a.online['path_to_model']))
                                                                       # print("Normalised B: " + str(params_b.online['path_to_model']))

    return mutable_a == mutable_b



#!/usr/bin/env python
import numpy as np
import numpy.testing as npt
import os
from caiman.base.movies import load, load_iter
from caiman.paths import caiman_datadir


def test_load_iter():
    fname = os.path.join(caiman_datadir(), 'example_movies', 'demoMovie.tif')
    for subindices in (None, slice(100, None, 2)):
        m = load_iter(fname, subindices=subindices)
        S = 0
        while True:
            try:
                S += np.sum(next(m))
            except StopIteration:
                break
        npt.assert_allclose(S, load(fname, subindices=subindices).sum(), rtol=1e-6)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" A set of utilities, mostly for post-processing and visualization

We put arrays on disk as raw bytes, extending along the first dimension.
Alongside each array x we ensure the value x.dtype which stores the string
description of the array's dtype.

See Also:
------------

@url
.. image::
@author  epnev
"""
# \package caiman/dource_ectraction/cnmf
# \version   1.0
# \copyright GNU General Public License v2.0
# \date Created on Sat Sep 12 15:52:53 2015

from builtins import str
from builtins import range
from past.utils import old_div

import cv2
import h5py
import logging
import numpy as np
import os
import pylab as pl
import scipy
from scipy.sparse import spdiags, issparse, csc_matrix, csr_matrix
import scipy.ndimage.morphology as morph
from skimage.feature.peak import _get_high_intensity_peaks
import tifffile
from typing import List

from .initialization import greedyROI
from ...base.rois import com
from ...mmapping import parallel_dot_product, load_memmap
from ...cluster import extract_patch_coordinates
from ...utils.stats import df_percentile


def decimation_matrix(dims, sub):
    D = np.prod(dims)
    if sub == 2 and D <= 10000:  # faster for small matrices
        ind = np.arange(D) // 2 - \
            np.arange(dims[0], dims[0] + D) // (dims[0] * 2) * (dims[0] // 2) - \
            (dims[0] % 2) * (np.arange(D) % (2 * dims[0]) > dims[0]) * (np.arange(1, 1 + D) % 2)
    else:
        def create_decimation_matrix_bruteforce(dims, sub):
            dims_ds = tuple(1 + (np.array(dims) - 1) // sub)
            d_ds = np.prod(dims_ds)
            ds_matrix = np.eye(d_ds)
            ds_matrix = np.repeat(np.repeat(
                ds_matrix.reshape((d_ds,) + dims_ds, order='F'), sub, 1),
                sub, 2)[:, :dims[0], :dims[1]].reshape((d_ds, -1), order='F')
            ds_matrix /= ds_matrix.sum(1)[:, None]
            ds_matrix = csc_matrix(ds_matrix, dtype=np.float32)
            return ds_matrix
        tmp = create_decimation_matrix_bruteforce((dims[0], sub), sub).indices
        ind = np.concatenate([tmp] * (dims[1] // sub + 1))[:D] + \
            np.arange(D) // (dims[0] * sub) * ((dims[0] - 1) // sub + 1)
    data = 1. / np.unique(ind, return_counts=True)[1][ind]
    return csc_matrix((data, ind, np.arange(1 + D)), dtype=np.float32)


def peak_local_max(image, min_distance=1, threshold_abs=None,
                   threshold_rel=None, exclude_border=True, indices=True,
                   num_peaks=np.inf, footprint=None):
    """Find peaks in an image as coordinate list or boolean mask.

    Adapted from skimage to use opencv for speed.
    Replaced scipy.ndimage.maximum_filter by cv2.dilate.

    Peaks are the local maxima in a region of `2 * min_distance + 1`
    (i.e. peaks are separated by at least `min_distance`).

    If peaks are flat (i.e. multiple adjacent pixels have identical
    intensities), the coordinates of all such pixels are returned.

    If both `threshold_abs` and `threshold_rel` are provided, the maximum
    of the two is chosen as the minimum intensity threshold of peaks.

    Parameters
    ----------
    image : ndarray
        Input image.
    min_distance : int, optional
        Minimum number of pixels separating peaks in a region of `2 *
        min_distance + 1` (i.e. peaks are separated by at least
        `min_distance`).
        To find the maximum number of peaks, use `min_distance=1`.
    threshold_abs : float, optional
        Minimum intensity of peaks. By default, the absolute threshold is
        the minimum intensity of the image.
    threshold_rel : float, optional
        Minimum intensity of peaks, calculated as `max(image) * threshold_rel`.
    exclude_border : int, optional
        If nonzero, `exclude_border` excludes peaks from
        within `exclude_border`-pixels of the border of the image.
    indices : bool, optional
        If True, the output will be an array representing peak
        coordinates.  If False, the output will be a boolean array shaped as
        `image.shape` with peaks present at True elements.
    num_peaks : int, optional
        Maximum number of peaks. When the number of peaks exceeds `num_peaks`,
        return `num_peaks` peaks based on highest peak intensity.
    footprint : ndarray of bools, optional
        If provided, `footprint == 1` represents the local region within which
        to search for peaks at every point in `image`.  Overrides
        `min_distance` (also for `exclude_border`).

    Returns
    -------
    output : ndarray or ndarray of bools

        * If `indices = True`  : (row, column, ...) coordinates of peaks.
        * If `indices = False` : Boolean array shaped like `image`, with peaks
          represented by True values.

    Notes
    -----
    The peak local maximum function returns the coordinates of local peaks
    (maxima) in an image. A maximum filter is used for finding local maxima.
    This operation dilates the original image. After comparison of the dilated
    and original image, this function returns the coordinates or a mask of the
    peaks where the dilated image equals the original image.

    Examples
    --------
    >>> img1 = np.zeros((7, 7))
    >>> img1[3, 4] = 1
    >>> img1[3, 2] = 1.5
    >>> img1
    array([[ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ,  0. ],
           [ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ,  0. ],
           [ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ,  0. ],
           [ 0. ,  0. ,  1.5,  0. ,  1. ,  0. ,  0. ],
           [ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ,  0. ],
           [ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ,  0. ],
           [ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ,  0. ]])

    >>> peak_local_max(img1, min_distance=1)
    array([[3, 4],
           [3, 2]])

    >>> peak_local_max(img1, min_distance=2)
    array([[3, 2]])

    >>> img2 = np.zeros((20, 20, 20))
    >>> img2[10, 10, 10] = 1
    >>> peak_local_max(img2, exclude_border=0)
    array([[10, 10, 10]])

    """
    if type(exclude_border) == bool:
        exclude_border = min_distance if exclude_border else 0

    out = np.zeros_like(image, dtype=np.bool)

    if np.all(image == image.flat[0]):
        if indices is True:
            return np.empty((0, 2), np.int)
        else:
            return out

    # Non maximum filter
    if footprint is not None:
        # image_max = ndi.maximum_filter(image, footprint=footprint,
        #                                mode='constant')
        image_max = cv2.dilate(image, footprint=footprint, iterations=1)
    else:
        size = 2 * min_distance + 1
        # image_max = ndi.maximum_filter(image, size=size, mode='constant')
        image_max = cv2.dilate(image, cv2.getStructuringElement(
            cv2.MORPH_RECT, (size, size)), iterations=1)
    mask = image == image_max

    if exclude_border:
        # zero out the image borders
        for i in range(mask.ndim):
            mask = mask.swapaxes(0, i)
            remove = (footprint.shape[i] if footprint is not None
                      else 2 * exclude_border)
            mask[:remove // 2] = mask[-remove // 2:] = False
            mask = mask.swapaxes(0, i)

    # find top peak candidates above a threshold
    thresholds = []
    if threshold_abs is None:
        threshold_abs = image.min()
    thresholds.append(threshold_abs)
    if threshold_rel is not None:
        thresholds.append(threshold_rel * image.max())
    if thresholds:
        mask &= image > max(thresholds)

    # Select highest intensities (num_peaks)
    coordinates = _get_high_intensity_peaks(image, mask, num_peaks)

    if indices is True:
        return coordinates
    else:
        nd_indices = tuple(coordinates.T)
        out[nd_indices] = True
        return out


def dict_compare(d1, d2):
    d1_keys = set(d1.keys())
    d2_keys = set(d2.keys())
    intersect_keys = d1_keys.intersection(d2_keys)
    added = d1_keys - d2_keys
    removed = d2_keys - d1_keys
    modified = {o : (d1[o], d2[o]) for o in intersect_keys if np.any(d1[o] != d2[o])}
    same = set(o for o in intersect_keys if np.all(d1[o] == d2[o]))
    return added, removed, modified, same


def computeDFF_traces(Yr, A, C, bl, quantileMin=8, frames_window=200):
    extract_DF_F(Yr, A, C, bl, quantileMin, frames_window)


def extract_DF_F(Yr, A, C, bl, quantileMin=8, frames_window=200, block_size=400, dview=None):
    """ Compute DFF function from cnmf output.

     Disclaimer: it might be memory inefficient

    Args:
        Yr: ndarray (2D)
            movie pixels X time

        A: scipy.sparse.coo_matrix
            spatial components (from cnmf cnm.A)

        C: ndarray
            temporal components (from cnmf cnm.C)

        bl: ndarray
            baseline for each component (from cnmf cnm.bl)

        quantile_min: float
            quantile minimum of the

        frames_window: int
            number of frames for running quantile

    Returns:
        Cdf:
            the computed Calcium acitivty to the derivative of f

    See Also:
        ..image::docs/img/onlycnmf.png
    """
    nA = np.array(np.sqrt(A.power(2).sum(0)).T)
    A = scipy.sparse.coo_matrix(A / nA.T)
    C = C * nA
    bl = (bl * nA.T).squeeze()
    nA = np.array(np.sqrt(A.power(2).sum(0)).T)

    T = C.shape[-1]
    if 'memmap' in str(type(Yr)):
        if block_size >= 500:
            print('Forcing single thread for memory issues')
            dview_res = None
        else:
            print('Using thread. If memory issues set block_size larger than 500')
            dview_res = dview

        AY = parallel_dot_product(Yr, A, dview=dview_res, block_size=block_size,
                                  transpose=True).T
    else:
        AY = A.T.dot(Yr)

    bas_val = bl[None, :]
    Bas = np.repeat(bas_val, T, 0).T
    AA = A.T.dot(A)
    AA.setdiag(0)
    Cf = (C - Bas) * (nA**2)
    C2 = AY - AA.dot(C)

    if frames_window is None or frames_window > T:
        Df = np.percentile(C2, quantileMin, axis=1)
        C_df = Cf / Df[:, None]

    else:
        Df = scipy.ndimage.percentile_filter(
            C2, quantileMin, (frames_window, 1))
        C_df = Cf / Df

    return C_df

def detrend_df_f(A, b, C, f, YrA=None, quantileMin=8, frames_window=500, 
                 flag_auto=True, use_fast=False, detrend_only=False):
    """ Compute DF/F signal without using the original data.
    In general much faster than extract_DF_F

    Args:
        A: scipy.sparse.csc_matrix
            spatial components (from cnmf cnm.A)

        b: ndarray
            spatial background components

        C: ndarray
            temporal components (from cnmf cnm.C)

        f: ndarray
            temporal background components

        YrA: ndarray
            residual signals

        quantile_min: float
            quantile used to estimate the baseline (values in [0,100])

        frames_window: int
            number of frames for computing running quantile

        flag_auto: bool
            flag for determining quantile automatically

        use_fast: bool
            flag for u´sing approximate fast percentile filtering

        detrend_only: bool (False)
            flag for only subtracting baseline and not normalizing by it.
            Used in 1p data processing where baseline fluorescence cannot be
            determined.

    Returns:
        F_df:
            the computed Calcium acitivty to the derivative of f
    """

    if C is None:
        logging.warning("There are no components for DF/F extraction!")
        return None
    
    if b is None or f is None:
        b = np.zeros((A.shape[0], 1))
        f = np.zeros((1, C.shape[1]))
        logging.warning("Background components not present. Results should" +
                        " not be interpreted as DF/F normalized but only" +
                        " as detrended.")
        detrend_only = True
    if 'csc_matrix' not in str(type(A)):
        A = scipy.sparse.csc_matrix(A)
    if 'array' not in str(type(b)):
        b = b.toarray()
    if 'array' not in str(type(C)):
        C = C.toarray()
    if 'array' not in str(type(f)):
        f = f.toarray()

    nA = np.sqrt(np.ravel(A.power(2).sum(axis=0)))
    nA_mat = scipy.sparse.spdiags(nA, 0, nA.shape[0], nA.shape[0])
    nA_inv_mat = scipy.sparse.spdiags(1. / nA, 0, nA.shape[0], nA.shape[0])
    A = A * nA_inv_mat
    C = nA_mat * C
    if YrA is not None:
        YrA = nA_mat * YrA

    F = C + YrA if YrA is not None else C
    B = A.T.dot(b).dot(f)
    T = C.shape[-1]

    if flag_auto:
        data_prct, val = df_percentile(F[:, :frames_window], axis=1)
        if frames_window is None or frames_window > T:
            Fd = np.stack([np.percentile(f, prctileMin) for f, prctileMin in
                           zip(F, data_prct)])
            Df = np.stack([np.percentile(f, prctileMin) for f, prctileMin in
                           zip(B, data_prct)])
            if not detrend_only:
                F_df = (F - Fd[:, None]) / (Df[:, None] + Fd[:, None])
            else:
                F_df = F - Fd[:, None]
        else:
            if use_fast:
                Fd = np.stack([fast_prct_filt(f, level=prctileMin,
                                              frames_window=frames_window) for
                               f, prctileMin in zip(F, data_prct)])
                Df = np.stack([fast_prct_filt(f, level=prctileMin,
                                              frames_window=frames_window) for
                               f, prctileMin in zip(B, data_prct)])
            else:
                Fd = np.stack([scipy.ndimage.percentile_filter(
                    f, prctileMin, (frames_window)) for f, prctileMin in
                    zip(F, data_prct)])
                Df = np.stack([scipy.ndimage.percentile_filter(
                    f, prctileMin, (frames_window)) for f, prctileMin in
                    zip(B, data_prct)])
            if not detrend_only:
                F_df = (F - Fd) / (Df + Fd)
            else:
                F_df = F - Fd
    else:
        if frames_window is None or frames_window > T:
            Fd = np.percentile(F, quantileMin, axis=1)
            Df = np.percentile(B, quantileMin, axis=1)
            if not detrend_only:
                F_df = (F - Fd[:, None]) / (Df[:, None] + Fd[:, None])
            else:
                F_df = F - Fd[:, None]
        else:
            Fd = scipy.ndimage.percentile_filter(
                F, quantileMin, (frames_window, 1))
            Df = scipy.ndimage.percentile_filter(
                B, quantileMin, (frames_window, 1))
            if not detrend_only:
                F_df = (F - Fd) / (Df + Fd)
            else:
                F_df = F - Fd

    return F_df

def fast_prct_filt(input_data, level=8, frames_window=1000):
    """
    Fast approximate percentage filtering
    """

    data = np.atleast_2d(input_data).copy()
    T = np.shape(data)[-1]
    downsampfact = frames_window

    elm_missing = int(np.ceil(T * 1.0 / downsampfact)
                      * downsampfact - T)
    padbefore = int(np.floor(elm_missing / 2.))
    padafter = int(np.ceil(elm_missing / 2.))
    tr_tmp = np.pad(data.T, ((padbefore, padafter), (0, 0)), mode='reflect')
    numFramesNew, num_traces = np.shape(tr_tmp)
    #% compute baseline quickly

    tr_BL = np.reshape(tr_tmp, (downsampfact, int(numFramesNew / downsampfact),
                                num_traces), order='F')

    tr_BL = np.percentile(tr_BL, level, axis=0)
    tr_BL = scipy.ndimage.zoom(np.array(tr_BL, dtype=np.float32),
                               [downsampfact, 1], order=3, mode='nearest',
                               cval=0.0, prefilter=True)

    if padafter == 0:
        data -= tr_BL.T
    else:
        data -= tr_BL[padbefore:-padafter].T

    return data.squeeze()
#%%
def detrend_df_f_auto(A, b, C, f, dims=None, YrA=None, use_annulus = True, 
                      dist1 = 7, dist2 = 5, frames_window=1000, 
                      use_fast = False):
    """
    Compute DF/F using an automated level of percentile filtering based on
    kernel density estimation.

    Args:
        A: scipy.sparse.csc_matrix
            spatial components (from cnmf cnm.A)

        b: ndarray
            spatial backgrounds

        C: ndarray
            temporal components (from cnmf cnm.C)

        f: ndarray
            temporal background components

        YrA: ndarray
            residual signals

        frames_window: int
            number of frames for running quantile

        use_fast: bool
            flag for using fast approximate percentile filtering

    Returns:
        F_df:
            the computed Calcium acitivty to the derivative of f
    """

    if 'csc_matrix' not in str(type(A)):
        A = scipy.sparse.csc_matrix(A)
    if 'array' not in str(type(b)):
        b = b.toarray()
    if 'array' not in str(type(C)):
        C = C.toarray()
    if 'array' not in str(type(f)):
        f = f.toarray()

    nA = np.sqrt(np.ravel(A.power(2).sum(axis=0)))
    nA_mat = scipy.sparse.spdiags(nA, 0, nA.shape[0], nA.shape[0])
    nA_inv_mat = scipy.sparse.spdiags(1. / nA, 0, nA.shape[0], nA.shape[0])
    A = A * nA_inv_mat
    C = nA_mat * C
    if YrA is not None:
        YrA = nA_mat * YrA

    F = C + YrA if YrA is not None else C
    K = A.shape[-1]
    A_ann = A.copy()

    if use_annulus:
        dist1 = 7
        dist2 = 5
        X, Y = np.meshgrid(np.arange(-dist1, dist1), np.arange(-dist1, dist1))
        R = np.sqrt(X**2+Y**2)
        R[R > dist1] = 0
        R[R < dist2] = 0
        R = R.astype('bool')

        for k in range(K):
            a = A[:, k].toarray().reshape(dims, order='F') > 0
            a2 = np.bitwise_xor(morph.binary_dilation(a, R), a)
            a2 = a2.astype(float).flatten(order='F')
            a2 /= np.sqrt(a2.sum())
            a2 = scipy.sparse.csc_matrix(a2)
            A_ann[:, k] = a2.T

    B = A_ann.T.dot(b).dot(f)
    T = C.shape[-1]

    data_prct, val = df_percentile(F[:, :frames_window], axis=1)

    if frames_window is None or frames_window > T:
        Fd = np.stack([np.percentile(f, prctileMin) for f, prctileMin in
                       zip(F, data_prct)])
        Df = np.stack([np.percentile(f, prctileMin) for f, prctileMin in
                       zip(B, data_prct)])
        F_df = (F - Fd[:, None]) / (Df[:, None] + Fd[:, None])
    else:
        if use_fast:
            Fd = np.stack([fast_prct_filt(f, level=prctileMin,
                                          frames_window=frames_window) for
                           f, prctileMin in zip(F, data_prct)])
            Df = np.stack([fast_prct_filt(f, level=prctileMin,
                                          frames_window=frames_window) for
                           f, prctileMin in zip(B, data_prct)])
        else:
            Fd = np.stack([scipy.ndimage.percentile_filter(
                f, prctileMin, (frames_window)) for f, prctileMin in
                zip(F, data_prct)])
            Df = np.stack([scipy.ndimage.percentile_filter(
                f, prctileMin, (frames_window)) for f, prctileMin in
                zip(B, data_prct)])
        F_df = (F - Fd) / (Df + Fd)

    return F_df

#%%


def manually_refine_components(Y, xxx_todo_changeme, A, C, Cn, thr=0.9, display_numbers=True,
                               max_number=None, cmap=None, **kwargs):
    """Plots contour of spatial components
     against a background image and allows to interactively add novel components by clicking with mouse

     Args:
         Y: ndarray
                   movie in 2D

         (dx,dy): tuple
                   dimensions of the square used to identify neurons (should be set to the galue of gsiz)

         A:   np.ndarray or sparse matrix
                   Matrix of Spatial components (d x K)

         Cn:  np.ndarray (2D)
                   Background image (e.g. mean, correlation)

         thr: scalar between 0 and 1
                   Energy threshold for computing contours (default 0.995)

         display_number:     Boolean
                   Display number of ROIs if checked (default True)

         max_number:    int
                   Display the number for only the first max_number components (default None, display all numbers)

         cmap:     string
                   User specifies the colormap (default None, default colormap)

     Returns:
         A: np.ndarray
             matrix A os estimated spatial component contributions

         C: np.ndarray
             array of estimated calcium traces
    """
    (dx, dy) = xxx_todo_changeme
    if issparse(A):
        A = np.array(A.todense())
    else:
        A = np.array(A)

    d1, d2 = np.shape(Cn)
    d, nr = np.shape(A)
    if max_number is None:
        max_number = nr

    x, y = np.mgrid[0:d1:1, 0:d2:1]

    pl.imshow(Cn, interpolation=None, cmap=cmap)
    cm = com(A, d1, d2)

    Bmat = np.zeros((np.minimum(nr, max_number), d1, d2))
    for i in range(np.minimum(nr, max_number)):
        indx = np.argsort(A[:, i], axis=None)[::-1]
        cumEn = np.cumsum(A[:, i].flatten()[indx]**2)
        cumEn /= cumEn[-1]
        Bvec = np.zeros(d)
        Bvec[indx] = cumEn
        Bmat[i] = np.reshape(Bvec, np.shape(Cn), order='F')

    T = np.shape(Y)[-1]

    pl.close()
    fig = pl.figure()
    ax = pl.gca()
    ax.imshow(Cn, interpolation=None, cmap=cmap,
              vmin=np.percentile(Cn[~np.isnan(Cn)], 1), vmax=np.percentile(Cn[~np.isnan(Cn)], 99))
    for i in range(np.minimum(nr, max_number)):
        pl.contour(y, x, Bmat[i], [thr])

    if display_numbers:
        for i in range(np.minimum(nr, max_number)):
            ax.text(cm[i, 1], cm[i, 0], str(i + 1))

    A3 = np.reshape(A, (d1, d2, nr), order='F')
    while True:
        pts = fig.ginput(1, timeout=0)

        if pts != []:
            print(pts)
            xx, yy = np.round(pts[0]).astype(np.int)
            coords_y = np.array(list(range(yy - dy, yy + dy + 1)))
            coords_x = np.array(list(range(xx - dx, xx + dx + 1)))
            coords_y = coords_y[(coords_y >= 0) & (coords_y < d1)]
            coords_x = coords_x[(coords_x >= 0) & (coords_x < d2)]
            a3_tiny = A3[coords_y[0]:coords_y[-1] +
                         1, coords_x[0]:coords_x[-1] + 1, :]
            y3_tiny = Y[coords_y[0]:coords_y[-1] +
                        1, coords_x[0]:coords_x[-1] + 1, :]

            dy_sz, dx_sz = np.shape(a3_tiny)[:-1]
            y2_tiny = np.reshape(y3_tiny, (dx_sz * dy_sz, T), order='F')
            a2_tiny = np.reshape(a3_tiny, (dx_sz * dy_sz, nr), order='F')
            y2_res = y2_tiny - a2_tiny.dot(C)
            y3_res = np.reshape(y2_res, (dy_sz, dx_sz, T), order='F')
            a__, c__, center__, b_in__, f_in__ = greedyROI(
                y3_res, nr=1, gSig=[np.floor(old_div(dx_sz, 2)), np.floor(old_div(dy_sz, 2))], gSiz=[dx_sz, dy_sz])

            a_f = np.zeros((d, 1))
            idxs = np.meshgrid(coords_y, coords_x)
            a_f[np.ravel_multi_index(
                idxs, (d1, d2), order='F').flatten()] = a__

            A = np.concatenate([A, a_f], axis=1)
            C = np.concatenate([C, c__], axis=0)
            indx = np.argsort(a_f, axis=None)[::-1]
            cumEn = np.cumsum(a_f.flatten()[indx]**2)
            cumEn /= cumEn[-1]
            Bvec = np.zeros(d)
            Bvec[indx] = cumEn
            bmat = np.reshape(Bvec, np.shape(Cn), order='F')
            pl.contour(y, x, bmat, [thr])
            pl.pause(.01)

        elif pts == []:
            break

        nr += 1
        A3 = np.reshape(A, (d1, d2, nr), order='F')

    return A, C

def app_vertex_cover(A):
    """ Finds an approximate vertex cover for a symmetric graph with adjacency matrix A.

    Args:
        A:  boolean 2d array (K x K)
            Adjacency matrix. A is boolean with diagonal set to 0

    Returns:
        L:   A vertex cover of A

    Authors:
    Eftychios A. Pnevmatikakis, Simons Foundation, 2015
    """

    L = []
    while A.any():
        nz = np.nonzero(A)[0]          # find non-zero edges
        u = nz[np.random.randint(0, len(nz))]
        A[u, :] = False
        A[:, u] = False
        L.append(u)

    return np.asarray(L)


def update_order(A, new_a=None, prev_list=None, method='greedy'):
    '''Determines the update order of the temporal components given the spatial
    components by creating a nest of random approximate vertex covers

     Args:
         A:    np.ndarray
              matrix of spatial components (d x K)
         new_a: sparse array
              spatial component that is added, in order to efficiently update the orders in online scenarios
         prev_list: list of list
              orders from previous iteration, you need to pass if new_a is not None

     Returns:
         O:  list of sets
             list of subsets of components. The components of each subset can be updated in parallel
         lo: list
             length of each subset

    Written by Eftychios A. Pnevmatikakis, Simons Foundation, 2015
    '''
    K = np.shape(A)[-1]
    if new_a is None and prev_list is None:

        if method is 'greedy':
            prev_list, count_list = update_order_greedy(A, flag_AA=False)
        else:
            prev_list, count_list = update_order_random(A, flag_AA=False)
        return prev_list, count_list

    else:

        if new_a is None or prev_list is None:
            raise Exception(
                'In the online update order you need to provide both new_a and prev_list')

        counter = 0

        AA = A.T.dot(new_a)
        for group in prev_list:
            if AA[list(group)].sum() == 0:
                group.append(K)
                counter += 1
                break

        if counter == 0:
            if prev_list is not None:
                prev_list = list(prev_list)
                prev_list.append([K])

        count_list = [len(gr) for gr in prev_list]

        return prev_list, count_list

def order_components(A, C):
    """Order components based on their maximum temporal value and size

    Args:
        A:  sparse matrix (d x K)
            spatial components

        C:  matrix or np.ndarray (K x T)
            temporal components

    Returns:
        A_or:  np.ndarray
            ordered spatial components

        C_or:  np.ndarray
            ordered temporal components

        srt:   np.ndarray
            sorting mapping
    """

    A = np.array(A.todense())
    nA2 = np.sqrt(np.sum(A**2, axis=0))
    K = len(nA2)
    A = np.array(np.matrix(A) * spdiags(old_div(1, nA2), 0, K, K))
    nA4 = np.sum(A**4, axis=0)**0.25
    C = np.array(spdiags(nA2, 0, K, K) * np.matrix(C))
    mC = np.ndarray.max(np.array(C), axis=1)
    srt = np.argsort(nA4 * mC)[::-1]
    A_or = A[:, srt] * spdiags(nA2[srt], 0, K, K)
    C_or = spdiags(old_div(1., nA2[srt]), 0, K, K) * (C[srt, :])

    return A_or, C_or, srt

def update_order_random(A, flag_AA=True):
    """Determies the update order of temporal components using
    randomized partitions of non-overlapping components
    """

    K = np.shape(A)[-1]
    if flag_AA:
        AA = A.copy()
    else:
        AA = A.T.dot(A)

    AA.setdiag(0)
    F = (AA) > 0
    F = F.toarray()
    rem_ind = np.arange(K)
    O = []
    lo = []
    while len(rem_ind) > 0:
        L = np.sort(app_vertex_cover(F[rem_ind, :][:, rem_ind]))
        if L.size:
            ord_ind = set(rem_ind) - set(rem_ind[L])
            rem_ind = rem_ind[L]
        else:
            ord_ind = set(rem_ind)
            rem_ind = []

        O.append(ord_ind)
        lo.append(len(ord_ind))

    return O[::-1], lo[::-1]


def update_order_greedy(A, flag_AA=True):
    """Determines the update order of the temporal components

    this, given the spatial components using a greedy method
    Basically we can update the components that are not overlapping, in parallel

    Args:
        A:  sparse crc matrix
            matrix of spatial components (d x K)
        OR:
            A.T.dot(A) matrix (d x d) if flag_AA = true

        flag_AA: boolean (default true)

     Returns:
         parllcomp:   list of sets
             list of subsets of components. The components of each subset can be updated in parallel

         len_parrllcomp:  list
             length of each subset

    Author:
        Eftychios A. Pnevmatikakis, Simons Foundation, 2017
    """
    K = np.shape(A)[-1]
    parllcomp:List = []
    for i in range(K):
        new_list = True
        for ls in parllcomp:
            if flag_AA:
                if A[i, ls].nnz == 0:
                    ls.append(i)
                    new_list = False
                    break
            else:
                if (A[:, i].T.dot(A[:, ls])).nnz == 0:
                    ls.append(i)
                    new_list = False
                    break

        if new_list:
            parllcomp.append([i])
    len_parrllcomp = [len(ls) for ls in parllcomp]
    return parllcomp, len_parrllcomp
#%%


def compute_residuals(Yr_mmap_file, A_, b_, C_, f_, dview=None, block_size=1000, num_blocks_per_run=5):
    '''compute residuals from memory mapped file and output of CNMF
        Args:
            A_,b_,C_,f_:
                from CNMF

            block_size: int
                number of pixels processed together

            num_blocks_per_run: int
                nnumber of parallel blocks processes

        Returns:
            YrA: ndarray
                residuals per neuron
    '''
    if not ('sparse' in str(type(A_))):
        A_ = scipy.sparse.coo_matrix(A_)

    Ab = scipy.sparse.hstack((A_, b_)).tocsc()

    Cf = np.vstack((C_, f_))

    nA = np.ravel(Ab.power(2).sum(axis=0))

    if 'mmap' in str(type(Yr_mmap_file)):
        YA = parallel_dot_product(Yr_mmap_file, Ab, dview=dview, block_size=block_size,
                                  transpose=True, num_blocks_per_run=num_blocks_per_run) * scipy.sparse.spdiags(old_div(1., nA), 0, Ab.shape[-1], Ab.shape[-1])
    else:
        YA = (Ab.T.dot(Yr_mmap_file)).T * \
            spdiags(old_div(1., nA), 0, Ab.shape[-1], Ab.shape[-1])

    AA = ((Ab.T.dot(Ab)) * scipy.sparse.spdiags(old_div(1., nA),
                                                0, Ab.shape[-1], Ab.shape[-1])).tocsr()

    return (YA - (AA.T.dot(Cf)).T)[:, :A_.shape[-1]].T

def normalize_AC(A, C, YrA, b, f, neurons_sn):
    """ Normalize to unit norm A and b
    Args:
        A,C,Yr,b,f:
            outputs of CNMF
    """
    if 'sparse' in str(type(A)):
        nA = np.ravel(np.sqrt(A.power(2).sum(0)))
    else:
        nA = np.ravel(np.sqrt((A**2).sum(0)))

    if A is not None:
        A /= nA

    if C is not None:
        C = np.array(C)
        C *= nA[:, None]

    if YrA is not None:
        YrA = np.array(YrA)
        YrA *= nA[:, None]

    if b is not None:
        if issparse(b):
            nB = np.ravel(np.sqrt(b.power(2).sum(0)))
            b = csc_matrix(b)
            for k, i in enumerate(b.indptr[:-1]):
                b.data[i:b.indptr[k + 1]] /= nB[k]
        else:
            nB = np.ravel(np.sqrt((b**2).sum(0)))
            b = np.atleast_2d(b)
            b /= nB
        if issparse(f):
            f = csr_matrix(f)
            for k, i in enumerate(f.indptr[:-1]):
                f.data[i:f.indptr[k + 1]] *= nB[k]
        else:
            f = np.atleast_2d(f)
            f *= nB[:, np.newaxis]

    if neurons_sn is not None:
        neurons_sn *= nA

    return csc_matrix(A), C, YrA, b, f, neurons_sn


def get_file_size(file_name, var_name_hdf5='mov'):
    """ Computes the dimensions of a file or a list of files without loading
    it/them in memory. An exception is thrown if the files have FOVs with
    different sizes
        Args:
            file_name: str or list
                locations of file(s) in memory

            var_name_hdf5: 'str'
                if loading from hdf5 name of the variable to load

        Returns:
            dims: list
                dimensions of FOV

            T: list
                number of timesteps in each file
    """
    if isinstance(file_name, str):
        if os.path.exists(file_name):
            _, extension = os.path.splitext(file_name)[:2]
            extension = extension.lower()
            if extension == '.mat':
                byte_stream, file_opened = scipy.io.matlab.mio._open_file(file_name, appendmat=False)
                mjv, mnv = scipy.io.matlab.mio.get_matfile_version(byte_stream)
                if mjv == 2:
                    extension = '.h5'
            if extension == '.tif' or extension == '.tiff':
                tffl = tifffile.TiffFile(file_name)
                siz = tffl.series[0].shape
                T, dims = siz[0], siz[1:]
            elif extension == '.avi':
                cap = cv2.VideoCapture(file_name)
                dims = [0, 0]
                try:
                    T = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
                    dims[1] = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
                    dims[0] = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
                except():
                    print('Roll back to opencv 2')
                    T = int(cap.get(cv2.cv.CV_CAP_PROP_FRAME_COUNT))
                    dims[1] = int(cap.get(cv2.cv.CV_CAP_PROP_FRAME_WIDTH))
                    dims[0] = int(cap.get(cv2.cv.CV_CAP_PROP_FRAME_HEIGHT))
            elif extension == '.mmap':
                filename = os.path.split(file_name)[-1]
                Yr, dims, T = load_memmap(os.path.join(
                        os.path.split(file_name)[0], filename))
            elif extension in ('.h5', '.hdf5', '.nwb'):
                with h5py.File(file_name, "r") as f:
                    kk = list(f.keys())
                    if len(kk) == 1:
                        siz = f[kk[0]].shape
                    elif var_name_hdf5 in f:
                        if extension == '.nwb':
                            siz = f[var_name_hdf5]['data'].shape
                        else:
                            siz = f[var_name_hdf5].shape
                    else:
                        logging.error('The file does not contain a variable' +
                                      'named {0}'.format(var_name_hdf5))
                        raise Exception('Variable not found. Use one of the above')
                T, dims = siz[0], siz[1:]
            elif extension in ('.sbx'):
                from ...base.movies import loadmat_sbx
                info = loadmat_sbx(file_name[:-4]+ '.mat')['info']
                dims = tuple((info['sz']).astype(int))
                # Defining number of channels/size factor
                if info['channels'] == 1:
                    info['nChan'] = 2
                    factor = 1
                elif info['channels'] == 2:
                    info['nChan'] = 1
                    factor = 2
                elif info['channels'] == 3:
                    info['nChan'] = 1
                    factor = 2
            
                # Determine number of frames in whole file
                T = int(os.path.getsize(
                    file_name[:-4] + '.sbx') / info['recordsPerBuffer'] / info['sz'][1] * factor / 4 - 1)
                
            else:
                raise Exception('Unknown file type')
            dims = tuple(dims)
        else:
            raise Exception('File not found!')
    elif isinstance(file_name, tuple):
        from ...base.movies import load
        dims = load(file_name[0], var_name_hdf5=var_name_hdf5).shape
        T = len(file_name)

    elif isinstance(file_name, list):
        if len(file_name) == 1:
            dims, T = get_file_size(file_name[0], var_name_hdf5=var_name_hdf5)
        else:
            dims, T = zip(*[get_file_size(fn, var_name_hdf5=var_name_hdf5)
                for fn in file_name])
    else:
        raise Exception('Unknown input type')
    return dims, T


def fast_graph_Laplacian(mmap_file, dims, max_radius=10, kernel='heat',
                         dview=None, sigma=1, thr=0.05, p=10, normalize=True,
                         use_NN=False, rf=None, strides=None):
    """ Computes an approximate affinity maps and its graph Laplacian for all
    pixels. For each pixel it restricts its attention to a given radius around
    it.
        Args:
            mmap_file: str
                Memory mapped file in pixel first order

            max_radius: float
                Maximum radius around each pixel

            kernel: str {'heat', 'binary', 'cos'}
                type of kernel

            dview: dview object
                multiprocessing or ipyparallel object for parallelization

            sigma: float
                standard deviation of Gaussian (heat) kernel

            thr: float
                threshold for affinity matrix

            p: int
                number of neighbors

            normalize: bool
                normalize vectors before computing affinity

            use_NN: bool
                use only p nearest neighbors

        Returns:
            W: scipy.sparse.csr_matrix
                Graph affinity matrix

            D: scipy.sparse.spdiags
                Diagonal of affinity matrix
                
            L: scipy.sparse.csr_matrix
                Graph Laplacian matrix
    """
    Np = np.prod(np.array(dims))
    if rf is None:
        pars = []
        for i in range(Np):
            pars.append([i, mmap_file, dims, max_radius, kernel, sigma, thr,
                         p, normalize, use_NN])
        if dview is None:
            res = list(map(fast_graph_Laplacian_pixel, pars))
        else:
            res = dview.map(fast_graph_Laplacian_pixel, pars, chunksize=128)
        indptr = np.cumsum(np.array([0] + [len(r[0]) for r in res]))
        indeces = [item for sublist in res for item in sublist[0]]
        data = [item for sublist in res for item in sublist[1]]
        W = scipy.sparse.csr_matrix((data, indeces, indptr), shape=[Np, Np])
        D = scipy.sparse.spdiags(W.sum(0), 0, Np, Np)
        L = D - W
    else:
        indices, _ = extract_patch_coordinates(dims, rf, strides)
        pars = []
        for i in range(len(indices)):
            pars.append([mmap_file, indices[i], kernel, sigma, thr, p,
                         normalize, use_NN])
        if dview is None:
            res = list(map(fast_graph_Laplacian_patches, pars))
        else:
            res = dview.map(fast_graph_Laplacian_patches, pars)
        W = res
        D = [scipy.sparse.spdiags(w.sum(0), 0, w.shape[0], w.shape[0]) for w in W]
        L = [d - w for (d, w) in zip(W, D)]
    return W, D, L


def fast_graph_Laplacian_patches(pars):
    """ Computes the full graph affinity matrix on a patch. See 
    fast_graph_Laplacian above for definition of arguments.
    """
    mmap_file, indices, kernel, sigma, thr, p, normalize, use_NN = pars
    if type(mmap_file) not in {'str', 'list'}:
        Yind = mmap_file
    else:
        Y = load_memmap(mmap_file)[0]
        Yind = np.array(Y[indices])
    if normalize:
        Yind -= Yind.mean(1)[:, np.newaxis]
        Yind /= np.sqrt((Yind**2).sum(1)[:, np.newaxis])
        yf = np.ones((Yind.shape[0], 1))
    else:
        yf = (Yind**2).sum(1)[:, np.newaxis]
    yyt = Yind.dot(Yind.T)
    W = np.exp(-(yf + yf.T - 2*yyt)/sigma) if kernel.lower() == 'heat' else yyt
    W[W<thr] = 0
    if kernel.lower() == 'binary':
        W[W>0] = 1
    if use_NN:
        ind = np.argpartition(W, -p, axis=1)[:, :-p]
        for i in range(W.shape[0]):
            W[i, ind[i]] = 0
        W = scipy.sparse.csr_matrix(W)
        W = (W + W.T)/2
    return W
    
def fast_graph_Laplacian_pixel(pars):
    """ Computes the i-th row of the Graph affinity matrix. See 
    fast_graph_Laplacian above for definition of arguments.
    """
    i, mmap_file, dims, max_radius, kernel, sigma, thr, p, normalize, use_NN = pars 
    iy, ix = np.unravel_index(i, dims, order='F')
    xx = np.arange(0, dims[1]) - ix
    yy = np.arange(0, dims[0]) - iy
    [XX, YY] = np.meshgrid(xx, yy)
    R = np.sqrt(XX**2 + YY**2)
    R = R.flatten('F')
    indeces = np.where(R < max_radius)[0]
    Y = load_memmap(mmap_file)[0]
    Yind = np.array(Y[indeces])
    y = np.array(Y[i, :])
    if normalize:
        Yind -= Yind.mean(1)[:, np.newaxis]
        Yind /= np.sqrt((Yind**2).sum(1)[:, np.newaxis])
        y -= y.mean()
        y /= np.sqrt((y**2).sum())
    D = Yind - y
    if kernel.lower() == 'heat':
        w = np.exp(-np.sum(D**2, axis=1)/sigma)
    else: # kernel.lower() == 'cos':
        w = Yind.dot(y.T)

    w[w<thr] = 0
    if kernel.lower() == 'binary':
        w[w>0] = 1
    if use_NN:
        ind = np.argpartition(w, -p)[-p:]
    else:
        ind = np.where(w>0)[0]

    return indeces[ind].tolist(), w[ind].tolist()
    


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from . import atm
from . import spikepursuit
from .volparams import volparams

try:
    profile
except:
    def profile(a): return a


class VOLPY(object):
    """ Spike Detection in Voltage Imaging
        The general file class which is used to find spikes of voltage imaging.
        Its architecture is similar to the one of scikit-learn calling the function fit
        to run everything which is part of the structure of the class.
        The output will be recorded in self.estimates.
        In order to use VolPy within CaImAn, you must install Keras into your conda environment. 
        You can do this by activating your environment, and then issuing the command 
        "conda install -c conda-forge keras".
    """

    def __init__(self, n_processes, dview=None, doCrossVal=False, doGlobalSubtract=False,
                 contextSize=50, censorSize=12, nPC_bg=8, tau_lp=3, tau_pred=1,
                 sigmas=np.array([1, 1.5, 2]), nIter=5, localAlign=False, globalAlign=True,
                 highPassRegression=False, params=None):
        """
            n_processes: int
                number of processed used (if in parallel this controls memory usage)

            dview: Direct View object
                for parallelization pruposes when using ipyparallel

            doCrossVal: boolean
                whether to use cross validation to optimize regression regularization parameters

            doGlobalSubtract: boolean
                whether to subtract the signal which can be predicted by the entire video

            contextSize: int
                number of pixels surrounding the ROI to use as context

            censorSize: int
                number of pixels surrounding the ROI to censor from the background PCA; roughly
                the spatial scale of scattered/dendritic neural signals, in pixels

            nPC_bg: int
                number of principle components used for background subtraction

            tau_lp: int
                time window for lowpass filter (seconds); signals slower than this will be ignored

            tau_pred: int
                time window in seconds for high pass filtering to make predictor for regression

            sigmas: 1-d array
                spatial smoothing radius imposed on spatial filter

            nIter: int
                number of iterations alternating between estimating temporal and spatial filters

            localAlign: boolean

            globalAlign: boolean

            highPassRegression: boolean
                whether to regress on a high-passed version of the data.
                Slightly improves detection of spikes, but makes subthreshold unreliable
        """

        if params is None:
            self.params = volparams(doCrossVal=doCrossVal, doGlobalSubtract=doGlobalSubtract,
                                    contextSize=contextSize, censorSize=censorSize, nPC_bg=nPC_bg,
                                    tau_lp=tau_lp, tau_pred=tau_pred, sigmas=sigmas, nIter=nIter,
                                    localAlign=localAlign, globalAlign=globalAlign,
                                    highPassRegression=highPassRegression)
        else:
            self.params = params
            # params.set('patch', {'n_processes': n_processes})
        self.estimates = {}

    def fit(self, n_processes=None, dview=None):
        """Run the volspike function to detect spikes and save the result
        into self.estimates
        """
        results = []
        fnames = self.params.data['fnames']
        fr = self.params.data['fr']

        if self.params.volspike['method'] == 'SpikePursuit':
            volspike = spikepursuit.volspike
        elif self.params.volspike['method'] == 'atm':
            volspike = atm.volspike
    
   
        N = len(self.params.data['index'])
        times = int(np.ceil(N/n_processes))
        for j in range(times):
            if j < (times - 1):
                li = [k for k in range(j*n_processes, (j+1)*n_processes)]
            else:
                li = [k for k in range(j*n_processes, N )]
   
            args_in = []
            
            for i in li:
                ROIs = self.params.data['ROIs'][i]
                if self.params.data['weights'] is None:
                    weights = None
                else:
                    weights = self.params.data['weights'][i]
                args_in.append([fnames, fr, i, ROIs, weights, self.params.volspike])

            if 'multiprocessing' in str(type(dview)):
                results_part = dview.map_async(volspike, args_in).get(4294967)
            elif dview is not None:
                results_part = dview.map_sync(volspike, args_in)
            else:
                results_part = list(map(volspike, args_in))
                    
            results = results + results_part
        
        N = len(results)
        print(N)
        self.estimates['spikeTimes'] = [results[i]['spikeTimes'] for i in range(N)]
        self.estimates['trace'] = [results[i]['yFilt'] for i in range(N)]
        self.estimates['recons_signal'] = [results[i]['recons_signal'] for i in range(N)]
        self.estimates['spatialFilter'] = [results[i]['spatialFilter'] for i in range(N)]
        self.estimates['templates'] = [results[i]['templates'] for i in range(N)]
        self.estimates['cellN'] = [results[i]['cellN'] for i in range(N)]
        self.estimates['templates'] = [results[i]['templates'] for i in range(N)]
        self.estimates['snr'] = [results[i]['snr'] for i in range(N)]
        self.estimates['num_spikes'] = [results[i]['num_spikes'] for i in range(N)]
        self.estimates['passedLocalityTest'] = [results[i]['passedLocalityTest'] for i in range(N)]
        self.estimates['low_spk'] = [results[i]['low_spk'] for i in range(N)]
        self.estimates['weights'] = [results[i]['weights'] for i in range(N)]
        self.estimates['bwexp'] = [results[i]['bwexp'] for i in range(N)]
        

        return self




#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" functions related to the creation and management of the cluster

We put arrays on disk as raw bytes, extending along the first dimension.
Alongside each array x we ensure the value x.dtype which stores the data type.

@author andrea giovannucci
"""

# \package caiman
# \version   1.0
# \copyright GNU General Public License v2.0
# \date Created on Thu Oct 20 12:07:09 2016

from builtins import zip
from builtins import str
from builtins import map
from builtins import range

import glob
import ipyparallel
from ipyparallel import Client
import logging
import multiprocessing
from multiprocessing import Pool
import numpy as np
import os
import platform
import psutil
import shlex
import shutil
import subprocess
import sys
import time
from typing import Any, Dict, List, Optional, Tuple, Union

from .mmapping import load_memmap

logger = logging.getLogger(__name__)


def extract_patch_coordinates(dims: Tuple,
                              rf: Union[List, Tuple],
                              stride: Union[List[int], Tuple],
                              border_pix: int = 0,
                              indices=[slice(None)] * 2) -> Tuple[List, List]:
    """
    Partition the FOV in patches
    and return the indexed in 2D and 1D (flatten, order='F') formats

    Args:
        dims: tuple of int
            dimensions of the original matrix that will be divided in patches

        rf: tuple of int
            radius of receptive field, corresponds to half the size of the square patch

        stride: tuple of int
            degree of overlap of the patches
    """

    sl_start = [0 if sl.start is None else sl.start for sl in indices]
    sl_stop = [dim if sl.stop is None else sl.stop for (sl, dim) in zip(indices, dims)]
    sl_step = [1 for sl in indices]    # not used
    dims_large = dims
    dims = np.minimum(np.array(dims) - border_pix, sl_stop) - np.maximum(border_pix, sl_start)

    coords_flat = []
    shapes = []
    iters = [list(range(rf[i], dims[i] - rf[i], 2 * rf[i] - stride[i])) + [dims[i] - rf[i]] for i in range(len(dims))]

    coords = np.empty(list(map(len, iters)) + [len(dims)], dtype=np.object)
    for count_0, xx in enumerate(iters[0]):
        coords_x = np.arange(xx - rf[0], xx + rf[0] + 1)
        coords_x = coords_x[(coords_x >= 0) & (coords_x < dims[0])]
        coords_x += border_pix * 0 + np.maximum(sl_start[0], border_pix)

        for count_1, yy in enumerate(iters[1]):
            coords_y = np.arange(yy - rf[1], yy + rf[1] + 1)
            coords_y = coords_y[(coords_y >= 0) & (coords_y < dims[1])]
            coords_y += border_pix * 0 + np.maximum(sl_start[1], border_pix)

            if len(dims) == 2:
                idxs = np.meshgrid(coords_x, coords_y)

                coords[count_0, count_1] = idxs
                shapes.append(idxs[0].shape[::-1])

                coords_ = np.ravel_multi_index(idxs, dims_large, order='F')
                coords_flat.append(coords_.flatten())
            else:      # 3D data

                if border_pix > 0:
                    raise Exception(
                        'The parameter border pix must be set to 0 for 3D data since border removal is not implemented')

                for count_2, zz in enumerate(iters[2]):
                    coords_z = np.arange(zz - rf[2], zz + rf[2] + 1)
                    coords_z = coords_z[(coords_z >= 0) & (coords_z < dims[2])]
                    idxs = np.meshgrid(coords_x, coords_y, coords_z)
                    shps = idxs[0].shape
                    shapes.append([shps[1], shps[0], shps[2]])
                    coords[count_0, count_1, count_2] = idxs
                    coords_ = np.ravel_multi_index(idxs, dims, order='F')
                    coords_flat.append(coords_.flatten())

    for i, c in enumerate(coords_flat):
        assert len(c) == np.prod(shapes[i])

    return list(map(np.sort, coords_flat)), shapes


#%%
def apply_to_patch(mmap_file, shape: Tuple[Any, Any, Any], dview, rf, stride, function, *args,
                   **kwargs) -> Tuple[List, Any, Tuple]:
    """
    apply function to patches in parallel or not

    Args:
        mmap_file: Memory-mapped variable
            Variable handle, first thing returned by load_memmap()

        shape: tuple of three elements
            dimensions of the original movie across y, x, and time

        dview: ipyparallel view on client
            if None

        rf: int
            half-size of the square patch in pixel

        stride: int
            amount of overlap between patches

    Returns:
        results

    Raises:
        Exception 'Something went wrong'

    """
    # This function is (currently) only used in some demos and use_cases code. Unclear if it's still necessary,
    # and it's been broken for awhile.
    (_, d1, d2) = shape

    if not np.isscalar(rf):
        rf1, rf2 = rf
    else:
        rf1 = rf
        rf2 = rf

    if not np.isscalar(stride):
        stride1, stride2 = stride
    else:
        stride1 = stride
        stride2 = stride

    idx_flat, idx_2d = extract_patch_coordinates((d1, d2), rf=(rf1, rf2), stride=(stride1, stride2))

    shape_grid = tuple(np.ceil((d1 * 1. / (rf1 * 2 - stride1), d2 * 1. / (rf2 * 2 - stride2))).astype(np.int))
    if d1 <= rf1 * 2:
        shape_grid = (1, shape_grid[1])
    if d2 <= rf2 * 2:
        shape_grid = (shape_grid[0], 1)

    logger.debug("Shape of grid is " + str(shape_grid))

    args_in = []

    for id_f, id_2d in zip(idx_flat[:], idx_2d[:]):
        args_in.append((mmap_file.filename, id_f, id_2d, function, args, kwargs))

    logger.debug("Flat index is of length " + str(len(idx_flat)))
    if dview is not None:
        try:
            file_res = dview.map_sync(function_place_holder, args_in)
            dview.results.clear()

        except:
            raise Exception('Something went wrong')
        finally:
            logger.warn('You may think that it went well but reality is harsh') # TODO Figure out a better message
    else:

        file_res = list(map(function_place_holder, args_in))
    return file_res, idx_flat, shape_grid


#%%


def function_place_holder(args_in: Tuple) -> np.ndarray:
    #todo: todocument

    file_name, idx_, shapes, function, args, kwargs = args_in
    Yr, _, _ = load_memmap(file_name)
    Yr = Yr[idx_, :]
    Yr.filename = file_name
    _, T = Yr.shape
    Y = np.reshape(Yr, (shapes[1], shapes[0], T), order='F').transpose([2, 0, 1])
    [T, d1, d2] = Y.shape

    res_fun = function(Y, *args, **kwargs)
    if type(res_fun) is not tuple:

        if res_fun.shape == (d1, d2):
            logger.debug('** reshaping form 2D to 1D')
            res_fun = np.reshape(res_fun, d1 * d2, order='F')

    return res_fun


#%%


def start_server(slurm_script: str = None, ipcluster: str = "ipcluster", ncpus: int = None) -> None:
    """
    programmatically start the ipyparallel server

    Args:
        ncpus: int
            number of processors

        ipcluster : str
            ipcluster binary file name; requires 4 path separators on Windows. ipcluster="C:\\\\Anaconda3\\\\Scripts\\\\ipcluster.exe"
            Default: "ipcluster"
    """
    logger.info("Starting cluster...")
    if ncpus is None:
        ncpus = psutil.cpu_count()

    if slurm_script is None:

        if ipcluster == "ipcluster":
            subprocess.Popen("ipcluster start -n {0}".format(ncpus), shell=True, close_fds=(os.name != 'nt'))
        else:
            subprocess.Popen(shlex.split("{0} start -n {1}".format(ipcluster, ncpus)),
                             shell=True,
                             close_fds=(os.name != 'nt'))
        time.sleep(1.5)
        # Check that all processes have started
        client = ipyparallel.Client()
        time.sleep(1.5)
        while len(client) < ncpus:
            sys.stdout.write(".")                              # Give some visual feedback of things starting
            sys.stdout.flush()                                 # (de-buffered)
            time.sleep(0.5)
        logger.debug('Making sure everything is up and running')
        client.direct_view().execute('__a=1', block=True)      # when done on all, we're set to go
    else:
        shell_source(slurm_script)
        pdir, profile = os.environ['IPPPDIR'], os.environ['IPPPROFILE']
        logger.debug([pdir, profile])
        c = Client(ipython_dir=pdir, profile=profile)
        ee = c[:]
        ne = len(ee)
        logger.info(('Running on %d engines.' % (ne)))
        c.close()
        sys.stdout.write("start_server: done\n")


def shell_source(script: str) -> None:
    """ Run a source-style bash script, copy resulting env vars to current process. """
    # XXX This function is weird and maybe not a good idea. People easily might expect
    #     it to handle conditionals. Maybe just make them provide a key-value file
    #introduce echo to indicate the end of the output
    pipe = subprocess.Popen(". %s; env; echo 'FINISHED_CLUSTER'" % script, stdout=subprocess.PIPE, shell=True)

    env = dict()
    while True:
        line = pipe.stdout.readline().decode('utf-8').rstrip()
        if 'FINISHED_CLUSTER' in line:         # find the keyword set above to determine the end of the output stream
            break
        logger.debug("shell_source parsing line[" + str(line) + "]")
        lsp = str(line).split("=", 1)
        if len(lsp) > 1:
            env[lsp[0]] = lsp[1]

    os.environ.update(env)
    pipe.stdout.close()


def stop_server(ipcluster: str = 'ipcluster', pdir: str = None, profile: str = None, dview=None) -> None:
    """
    programmatically stops the ipyparallel server

    Args:
        ipcluster : str
            ipcluster binary file name; requires 4 path separators on Windows
            Default: "ipcluster"a

        pdir : Undocumented
        profile: Undocumented
        dview: Undocumented

    """
    if 'multiprocessing' in str(type(dview)):
        dview.terminate()
    else:
        logger.info("Stopping cluster...")
        try:
            pdir, profile = os.environ['IPPPDIR'], os.environ['IPPPROFILE']
            is_slurm = True
        except:
            logger.debug('stop_server: not a slurm cluster')
            is_slurm = False

        if is_slurm:
            if pdir is None and profile is None:
                pdir, profile = os.environ['IPPPDIR'], os.environ['IPPPROFILE']
            c = Client(ipython_dir=pdir, profile=profile)
            ee = c[:]
            ne = len(ee)
            logger.info(('Shutting down %d engines.' % (ne)))
            c.close()
            c.shutdown(hub=True)
            shutil.rmtree('profile_' + str(profile))
            try:
                shutil.rmtree('./log/')
            except:
                logger.info('creating log folder')     # FIXME Not what this means

            files = glob.glob('*.log')
            os.mkdir('./log')

            for fl in files:
                shutil.move(fl, './log/')

        else:
            if ipcluster == "ipcluster":
                proc = subprocess.Popen("ipcluster stop",
                                        shell=True,
                                        stderr=subprocess.PIPE,
                                        close_fds=(os.name != 'nt'))
            else:
                proc = subprocess.Popen(shlex.split(ipcluster + " stop"),
                                        shell=True,
                                        stderr=subprocess.PIPE,
                                        close_fds=(os.name != 'nt'))

            line_out = proc.stderr.readline()
            if b'CRITICAL' in line_out:
                logger.info("No cluster to stop...")
            elif b'Stopping' in line_out:
                st = time.time()
                logger.debug('Waiting for cluster to stop...')
                while (time.time() - st) < 4:
                    sys.stdout.write('.')
                    sys.stdout.flush()
                    time.sleep(1)
            else:
                logger.error(line_out)
                logger.error('**** Unrecognized syntax in ipcluster output, waiting for server to stop anyways ****')

            proc.stderr.close()

    logger.info("stop_cluster(): done")


#%%


def setup_cluster(backend: str = 'multiprocessing',
                  n_processes: int = None,
                  single_thread: bool = False,
                  ignore_preexisting: bool = False,
                  maxtasksperchild: int = None) -> Tuple[Any, Any, Optional[int]]:
    """Setup and/or restart a parallel cluster.
    Args:
        backend: str
            'multiprocessing' [alias 'local'], 'ipyparallel', and 'SLURM'
            ipyparallel and SLURM backends try to restart if cluster running.
            backend='multiprocessing' raises an exception if a cluster is running.
        ignore_preexisting: bool
            If True, ignores the existence of an already running multiprocessing
            pool, which is usually indicative of a previously-started CaImAn cluster

    Returns:
        c: ipyparallel.Client object; only used for ipyparallel and SLURM backends, else None
        dview: ipyparallel dview object, or for multiprocessing: Pool object
        n_processes: number of workers in dview. None means guess at number of machine cores.
    """

    if n_processes is None:
        if backend == 'SLURM':
            n_processes = np.int(os.environ.get('SLURM_NPROCS'))
        else:
            # roughly number of cores on your machine minus 1
            n_processes = np.maximum(np.int(psutil.cpu_count() - 1), 1)

    if single_thread:
        dview = None
        c = None
    else:
        sys.stdout.flush()

        if backend == 'SLURM':
            try:
                stop_server()
            except:
                logger.debug('Nothing to stop')
            slurm_script = '/mnt/home/agiovann/SOFTWARE/CaImAn/SLURM/slurmStart.sh' # FIXME: Make this a documented environment variable
            logger.info([str(n_processes), slurm_script])
            start_server(slurm_script=slurm_script, ncpus=n_processes)
            pdir, profile = os.environ['IPPPDIR'], os.environ['IPPPROFILE']
            logger.info([pdir, profile])
            c = Client(ipython_dir=pdir, profile=profile)
            dview = c[:]
        elif backend == 'ipyparallel':
            stop_server()
            start_server(ncpus=n_processes)
            c = Client()
            logger.info(f'Started ipyparallel cluster: Using {len(c)} processes')
            dview = c[:len(c)]

        elif (backend == 'multiprocessing') or (backend == 'local'):
            if len(multiprocessing.active_children()) > 0:
                if ignore_preexisting:
                    logger.warn('Found an existing multiprocessing pool. '
                                'This is often indicative of an already-running CaImAn cluster. '
                                'You have configured the cluster setup to not raise an exception.')
                else:
                    raise Exception(
                        'A cluster is already runnning. Terminate with dview.terminate() if you want to restart.')
            if (platform.system() == 'Darwin') and (sys.version_info > (3, 0)):
                try:
                    if 'kernel' in get_ipython().trait_names():        # type: ignore
                                                                       # If you're on OSX and you're running under Jupyter or Spyder,
                                                                       # which already run the code in a forkserver-friendly way, this
                                                                       # can eliminate some setup and make this a reasonable approach.
                                                                       # Otherwise, seting VECLIB_MAXIMUM_THREADS=1 or using a different
                                                                       # blas/lapack is the way to avoid the issues.
                                                                       # See https://github.com/flatironinstitute/CaImAn/issues/206 for more
                                                                       # info on why we're doing this (for now).
                        multiprocessing.set_start_method('forkserver', force=True)
                except:                                                # If we're not running under ipython, don't do anything.
                    pass
            c = None

            dview = Pool(n_processes, maxtasksperchild=maxtasksperchild)
        else:
            raise Exception('Unknown Backend')

    return c, dview, n_processes



"""
Mask R-CNN
Train on the segmentation of neurons.

Copyright (c) 2018 Matterport, Inc.
Licensed under the MIT License (see LICENSE for details)
Written by Waleed Abdulla

------------------------------------------------------------
"""

import os
import sys
import json
import datetime
import numpy as np
import skimage.draw

# Root directory of the project
ROOT_DIR = os.path.abspath("../../")

# Import Mask RCNN
sys.path.append(ROOT_DIR)  # To find local version of the library
from ..mrcnn.config import Config
from ..mrcnn import model as modellib, utils

# Path to trained weights file
COCO_WEIGHTS_PATH = os.path.join(ROOT_DIR, "mask_rcnn_coco.h5")

# Directory to save logs and model checkpoints, if not provided
# through the command line argument --logs
DEFAULT_LOGS_DIR = os.path.join(ROOT_DIR, "logs")

############################################################
#  Configurations
############################################################
class NeuronsConfig(Config):
    """Configuration for training on the nucleus segmentation dataset."""
    # Give the configuration a recognizable name
    NAME = "neurons"

    # Adjust depending on your GPU memory
    IMAGES_PER_GPU = 6

    # Number of classes (including background)
    NUM_CLASSES = 1 + 1  # Background + nucleus

    # Number of training and validation steps per epoch
    #STEPS_PER_EPOCH = (657 - len(VAL_IMAGE_IDS)) // IMAGES_PER_GPU
    #VALIDATION_STEPS = max(1, len(VAL_IMAGE_IDS) // IMAGES_PER_GPU)
    STEPS_PER_EPOCH = 100    
    
    # Don't exclude based on confidence. Since we have two classes
    # then 0.5 is the minimum anyway as it picks between nucleus and BG
    DETECTION_MIN_CONFIDENCE = 0

    # Backbone network architecture
    # Supported values are: resnet50, resnet101
    BACKBONE = "resnet50"

    # Input image resizing
    # Random crops of size 512x512
    IMAGE_RESIZE_MODE = "crop"
    IMAGE_MIN_DIM = 128
    #IMAGE_MIN_DIM = 512
    #IMAGE_MAX_DIM = 512
    #IMAGE_MIN_SCALE = 2.0

    # Length of square anchor side in pixels
    RPN_ANCHOR_SCALES = (16, 32, 64, 128, 256)   #(8, 16, 32, 64, 128) 

    # ROIs kept after non-maximum supression (training and inference)
    POST_NMS_ROIS_TRAINING = 1000
    POST_NMS_ROIS_INFERENCE = 2000

    # Non-max suppression threshold to filter RPN proposals.
    # You can increase this during training to generate more propsals.
    RPN_NMS_THRESHOLD = 0.9

    # How many anchors per image to use for RPN training
    RPN_TRAIN_ANCHORS_PER_IMAGE = 64

    # Image mean (RGB)
    MEAN_PIXEL = np.array([0,0,0])

    # If enabled, resizes instance masks to a smaller size to reduce
    # memory load. Recommended when using high-resolution images.
    USE_MINI_MASK = True
    MINI_MASK_SHAPE = (56, 56)  # (height, width) of the mini-mask

    # Number of ROIs per image to feed to classifier/mask heads
    # The Mask RCNN paper uses 512 but often the RPN doesn't generate
    # enough positive proposals to fill this and keep a positive:negative
    # ratio of 1:3. You can increase the number of proposals by adjusting
    # the RPN NMS threshold.
    TRAIN_ROIS_PER_IMAGE = 128

    # Maximum number of ground truth instances to use in one image
    MAX_GT_INSTANCES = 100

    # Max number of final detections per image
    DETECTION_MAX_INSTANCES = 100
    
    WEIGHT_DECAY = 0.0001
   
############################################################
#  Dataset
############################################################

class NeuronsDataset(utils.Dataset):

    def load_neurons(self, dataset_dir, subset):
        """Load a subset of the neurons dataset.
        dataset_dir: Root directory of the dataset.
        subset: Subset to load: train or val
        """
        # Add classes. We have only one class to add.
        self.add_class("neurons", 1, "neurons")

        # Train or validation dataset?
        assert subset in ["train", "val"]
        dataset_dir = os.path.join(dataset_dir, subset)

        # Load annotations
        # VGG Image Annotator (up to version 1.6) saves each image in the form:
        # { 'filename': '28503151_5b5b7ec140_b.jpg',
        #   'regions': {
        #       '0': {
        #           'region_attributes': {},
        #           'shape_attributes': {
        #               'all_points_x': [...],
        #               'all_points_y': [...],
        #               'name': 'polygon'}},
        #       ... more regions ...
        #   },
        #   'size': 100202
        # }
        # We mostly care about the x and y coordinates of each region
        # Note: In VIA 2.0, regions was changed from a dict to a list.
        """
        group = [i for i in sorted(os.listdir(dataset_dir)) if i[-4:]=='.npz']
        
        # Add images
        for a in group:
            fname = dataset_dir + '/' + a
            print(fname)
            import zipfile
            from caiman.base.rois import nf_read_roi
            with zipfile.ZipFile(fname[:-4]+'.zip') as zf:
                names = zf.namelist()
                coords = [nf_read_roi(zf.open(n))
                          for n in names]
                polygons = [{'name': 'polygon','all_points_x':i[:,1],'all_points_y':i[:,0]} for i in coords]
            # Get the x, y coordinaets of points of the polygons that make up
            # the outline of each object instance. These are stores in the
            # shape_attributes (see json format above)
            # The if condition is needed to support VIA versions 1.x and 2.x.
            
            # load_mask() needs the image size to convert polygons to masks.
            # Unfortunately, VIA doesn't include it in JSON, so we must read
            # the image. This is only managable since the dataset is tiny.
            image_path = os.path.join(dataset_dir, a)
            image = np.load(image_path)['arr_0']
            height, width = image.shape[:2]

            self.add_image(
                source = "neurons",
                image_id=a,  # use file name as a unique image id
                path=image_path,
                width=width, height=height,
                polygons=polygons)
                
         """
        group = [i for i in sorted(os.listdir(dataset_dir)) if 'mask' not in i]
        for a in group:
            fname = dataset_dir + '/' + a
            print(fname)
            polygons = np.load(fname[:-4]+'_mask.npz', allow_pickle=True)['arr_0']
            image_path = os.path.join(dataset_dir, a)
            image = np.load(image_path)['arr_0'].astype(np.float32)
            height, width = image.shape[:2]
            self.add_image(
                source = "neurons",
                image_id=a,  # use file name as a unique image id
                path=image_path,
                width=width, height=height,
                polygons=polygons)
        
       

    def load_mask(self, image_id):
        """Generate instance masks for an image.
       Returns:
        masks: A bool array of shape [height, width, instance count] with
            one mask per instance.
        class_ids: a 1D array of class IDs of the instance masks.
        """
        # If not a neuron dataset image, delegate to parent class.
        image_info = self.image_info[image_id]
        if image_info["source"] != "neurons":
            return super(self.__class__, self).load_mask(image_id)

        # Convert polygons to a bitmap mask of shape
        # [height, width, instance_count]
        info = self.image_info[image_id]
        mask = np.zeros([info["height"], info["width"], len(info["polygons"])],
                        dtype=np.uint8)
        for i, p in enumerate(info["polygons"]):
            # Get indexes of pixels inside the polygon and set them to 1
            rr, cc = skimage.draw.polygon(p['all_points_y'], p['all_points_x'])
            mask[rr, cc, i] = 1

        # Return mask, and array of class IDs of each instance. Since we have
        # one class ID only, we return an array of 1s
        return mask.astype(np.bool), np.ones([mask.shape[-1]], dtype=np.int32)

    def image_reference(self, image_id):
        """Return the path of the image."""
        info = self.image_info[image_id]
        if info["source"] == "neurons":
            return info["path"]
        else:
            super(self.__class__, self).image_reference(image_id)


def train(model):
    """Train the model."""
    # Training dataset.
    dataset_train = NeuronsDataset()
    dataset_train.load_neurons(args.dataset, "train")
    dataset_train.prepare()

    # Validation dataset
    dataset_val = NeuronsDataset()
    dataset_val.load_neurons(args.dataset, "val")
    dataset_val.prepare()

    # *** This training schedule is an example. Update to your needs ***
    # Since we're using a very small dataset, and starting from
    # COCO trained weights, we don't need to train too long. Also,
    # no need to train all layers, just the heads should do it.
    import imgaug.augmenters as iaa
    sometimes = lambda aug: iaa.Sometimes(0.5, aug)
    seldom = lambda aug: iaa.Sometimes(0.2, aug)
    
    augmentation = iaa.Sequential([iaa.Fliplr(0.5),
                                  iaa.Flipud(0.5),
                                  iaa.Affine(rotate=(-180, 180)),          
                                  #sometimes(iaa.GaussianBlur(sigma=(0, 0.25))),
                                  sometimes(iaa.Multiply((0.9,1.1))),
                                  sometimes(iaa.AdditiveGaussianNoise(loc=0, scale=(0.0, 0.05))),
                                  sometimes(iaa.Affine(shear=(-2,2))),
                                  sometimes(iaa.Affine(scale=(0.9, 1.1),
                                  translate_percent={"x": (-0.1, 0.1), "y": (-0.1, 0.1)}))],random_order=True)
    print("Train network heads")
    model.train(dataset_train, dataset_val,
                learning_rate=1e-2,
                epochs= 20,
                layers='heads',
                augmentation = augmentation)
    print("Train since resnet stage 4")
    model.train(dataset_train, dataset_val,
                learning_rate=1e-3,
                epochs= 40,
                layers='4+',
                augmentation = augmentation)


def color_splash(image, mask):
    """Apply color splash effect.
    image: RGB image [height, width, 3]
    mask: instance segmentation mask [height, width, instance count]

    Returns result image.
    """
    # Make a grayscale copy of the image. The grayscale copy still
    # has 3 RGB channels, though.
    gray = skimage.color.gray2rgb(skimage.color.rgb2gray(image)) * 255
    # Copy color pixels from the original color image where mask is set
    if mask.shape[-1] > 0:
        # We're treating all instances as one, so collapse the mask into one layer
        mask = (np.sum(mask, -1, keepdims=True) >= 1)
        splash = np.where(mask, image, gray).astype(np.uint8)
    else:
        splash = gray.astype(np.uint8)
    return splash


def detect_and_color_splash(model, image_path=None, video_path=None):
    assert image_path or video_path

    # Image or video?
    if image_path:
        # Run model detection and generate the color splash effect
        print("Running on {}".format(args.image))
        # Read image
        image = skimage.io.imread(args.image)
        # Detect objects
        r = model.detect([image], verbose=1)[0]
        # Color splash
        splash = color_splash(image, r['masks'])
        # Save output
        file_name = "splash_{:%Y%m%dT%H%M%S}.png".format(datetime.datetime.now())
        skimage.io.imsave(file_name, splash)
    elif video_path:
        import cv2
        # Video capture
        vcapture = cv2.VideoCapture(video_path)
        width = int(vcapture.get(cv2.CAP_PROP_FRAME_WIDTH))
        height = int(vcapture.get(cv2.CAP_PROP_FRAME_HEIGHT))
        fps = vcapture.get(cv2.CAP_PROP_FPS)

        # Define codec and create video writer
        file_name = "splash_{:%Y%m%dT%H%M%S}.avi".format(datetime.datetime.now())
        vwriter = cv2.VideoWriter(file_name,
                                  cv2.VideoWriter_fourcc(*'MJPG'),
                                  fps, (width, height))

        count = 0
        success = True
        while success:
            print("frame: ", count)
            # Read next image
            success, image = vcapture.read()
            if success:
                # OpenCV returns images as BGR, convert to RGB
                image = image[..., ::-1]
                # Detect objects
                r = model.detect([image], verbose=0)[0]
                # Color splash
                splash = color_splash(image, r['masks'])
                # RGB -> BGR to save image to video
                splash = splash[..., ::-1]
                # Add image to video writer
                vwriter.write(splash)
                count += 1
        vwriter.release()
    print("Saved to ", file_name)

############################################################
#  Training
############################################################

if __name__ == '__main__':
    import argparse

    # Parse command line arguments
    parser = argparse.ArgumentParser(
        description='Train Mask R-CNN to detect neurons.')
    parser.add_argument("command",
                        metavar="<command>",
                        help="'train' or 'splash'")
    parser.add_argument('--dataset', required=False,
                        metavar="/path/to/neuron/dataset/",
                        help='Directory of the Neurons dataset')
    parser.add_argument('--weights', required=True,
                        metavar="/path/to/weights.h5",
                        help="Path to weights .h5 file or 'coco'")
    parser.add_argument('--logs', required=False,
                        default=DEFAULT_LOGS_DIR,
                        metavar="/path/to/logs/",
                        help='Logs and checkpoints directory (default=logs/)')
    parser.add_argument('--image', required=False,
                        metavar="path or URL to image",
                        help='Image to apply the color splash effect on')
    parser.add_argument('--video', required=False,
                        metavar="path or URL to video",
                        help='Video to apply the color splash effect on')
    args = parser.parse_args()

    # Validate arguments
    if args.command == "train":
        assert args.dataset, "Argument --dataset is required for training"
    elif args.command == "splash":
        assert args.image or args.video,\
               "Provide --image or --video to apply color splash"

    print("Weights: ", args.weights)
    print("Dataset: ", args.dataset)
    print("Logs: ", args.logs)

    # Configurations
    if args.command == "train":
        config = NeuronsConfig()
    else:
        class InferenceConfig(NeuronsConfig):
            # Set batch size to 1 since we'll be running inference on
            # one image at a time. Batch size = GPU_COUNT * IMAGES_PER_GPU
            GPU_COUNT = 1
            IMAGES_PER_GPU = 1
            IMAGE_RESIZE_MODE = "none"
        config = InferenceConfig()
    config.display()

    # Create model
    if args.command == "train":
        model = modellib.MaskRCNN(mode="training", config=config,
                                  model_dir=args.logs)
    else:
        model = modellib.MaskRCNN(mode="inference", config=config,
                                  model_dir=args.logs)

    # Select weights file to load
    if args.weights.lower() == "coco":
        weights_path = COCO_WEIGHTS_PATH
        # Download weights file
        if not os.path.exists(weights_path):
            utils.download_trained_weights(weights_path)
    elif args.weights.lower() == "last":
        # Find last trained weights
        weights_path = model.find_last()
    elif args.weights.lower() == "imagenet":
        # Start from ImageNet trained weights
        weights_path = model.get_imagenet_weights()
    else:
        weights_path = args.weights

    # Load weights
    print("Loading weights ", weights_path)
    if args.weights.lower() == "coco":
        # Exclude the last layers because they require a matching
        # number of classes
        model.load_weights(weights_path, by_name=True, exclude=[
            "mrcnn_class_logits", "mrcnn_bbox_fc",
            "mrcnn_bbox", "mrcnn_mask"])
    else:
        model.load_weights(weights_path, by_name=True)

    # Train or evaluate
    if args.command == "train":
        train(model)
    elif args.command == "splash":
        detect_and_color_splash(model, image_path=args.image,
                                video_path=args.video)
    else:
        print("'{}' is not recognized. "
              "Use 'train' or 'splash'".format(args.command))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Apr 19 14:50:09 2019

@author: @caichangjia based on Matlab code provided by Kaspar Podgorski and Amrita Singh
"""
import numpy as np
import matplotlib.pyplot as plt
from skimage.morphology import dilation
from skimage.morphology import disk
from sklearn.linear_model import LinearRegression
from sklearn.linear_model import Ridge
from scipy import signal
from scipy import stats    
from scipy.sparse.linalg import svds
import cv2
from caiman.base.movies import movie
import caiman as cm


# %%
def volspike(pars):
    """ Main function for finding spikes of one single neuron with given ROI in
        voltage imaging. Using function denoiseSpikes to find spikes
        of one dimensional signal, using ridge regression to find the
        best spatial filters. Do these two steps iteratively to find
        best spike time.

        Args:
            pars: list
                fnames: str
                    name of the memory map file

                fr: int

                cellN: int
                    number of cell processing

                ROIs: 3-d array
                    all region of interests

                weights: 3-d array
                    spatial weights of different cells generated by previous data blocks as initialization

                args: dictionary

                    doCrossVal: boolean
                        whether to use cross validation to optimize regression regularization parameters

                    doGlobalSubtract: boolean
                        whether to subtract the signal which can be predicted by the entire video

                    contextSize: int
                        number of pixels surrounding the ROI to use as context

                    censorSize: int
                        number of pixels surrounding the ROI to censor from the background PCA; roughly
                        the spatial scale of scattered/dendritic neural signals, in pixels

                    nPC_bg: int
                        number of principle components used for background subtraction

                    tau_lp: int
                        time window for lowpass filter (seconds); signals slower than this will be ignored

                    tau_pred: int
                        time window in seconds for high pass filtering to make predictor for regression

                    sigmas: 1-d array
                        spatial smoothing radius imposed on spatial filter

                    nIter: int
                        number of iterations alternating between estimating temporal and spatial filters

                    localAlign: boolean

                    globalAlign: boolean

                    highPassRegression: boolean
                        whether to regress on a high-passed version of the data. Slightly improves detection of spikes,
                        but makes subthreshold unreliable
                        
                    use_ridge: boolean
                        whether use ridge regression for removing background signal. If not, the algorithm will use 
                        linear regression
                        
                    Ridge_bg_coef: float
                        regularization strength for ridge regression in background removal. 

        Returns:
            output: dictionary
                
                rawROI: dictionary
                    including the result after the first spike extraction
                    
                bwexp: 2-d array
                    expansion area of ROI, used for reconstructing signal
                    
                meanIM: 1-d array
                    trace after high-pass filter
                    
                num_spikes: list
                    number of spikes in each iteration
                    
                passedLocalityTest: boolean
                    False if the maximum of spatial filter is not in the initial ROI
                    
                snr: float
                    signal to noise ratio of the processed signal
                    
                y: 1-d array
                    processed signal of the video
                
                yFilt: 1-d array
                    spike train of the signal
                    
                ROIbw: 2-d array
                    region of interest
                    
                recons_signal: 1-d array
                    reconstructed signal of the neuron
                    
                spatialFilter: 2-d array
                    spatial filter of the neuron
                    
                falsePosRate: float
                    possibility of misclassify noise as real spikes

                detectionRate: float
                    possibility of real spikes being detected
                    
                template: 1-d array
                    spike template of the neuron
                    
                spikeTimes: 1-d array
                    spike time of the neuron
                    
                thresh: float
                    threshold of the signal
                    
                F0: 1-d array
                    initial signal 
                    
                dFF: 1-d array
                    scaled signal
                
                bg_pc: 2-d array
                    background principal components extracted from svd
                    
                low_spk: boolean
                    whether there are few spikes(less than 30) in the signal
                    
                weights: 2-d array
                    ridge regression coefficient for updating spatial filter
                                        
                cellN: int
                    index of cell            
                
    """
    fnames = pars[0]
    sampleRate = pars[1]
    cellN = pars[2]
    bw = pars[3]    
    weights_init = pars[4]    
    args = pars[5]
    
    print('Now processing cell number {0}'.format(cellN))
    
    doCrossVal = args['doCrossVal']
    doGlobalSubtract = args['doGlobalSubtract']
    contextSize = args['contextSize']
    censorSize = args['censorSize']
    nPC_bg = args['nPC_bg']
    tau_lp = args['tau_lp']
    tau_pred = args['tau_pred']
    sigmas = args['sigmas']
    nIter = args['nIter']
    localAlign = args['localAlign']
    globalAlign = args['globalAlign']
    highPassRegression = args['highPassRegression']
    use_Ridge = args['use_Ridge']
    Ridge_bg_coef = args['Ridge_bg_coef']
    windowLength = sampleRate * 0.02 # window length for spike templates
    output = {}
    output['rawROI'] = {}

    Yr, dims, T = cm.load_memmap(fnames)
    if bw.shape == dims:
        images = np.reshape(Yr.T, [T] + list(dims), order='F')
    elif bw.shape == dims[::-1]:
        images = np.reshape(Yr.T, [T] + list(dims), order='F').transpose([0, 2, 1])
    else:
        print('size of ROI and video does not accrod')

    # extract relevant region and align
    bwexp = dilation(bw, np.ones([contextSize, contextSize]), shift_x=True, shift_y=True)
    Xinds = np.where(np.any(bwexp > 0, axis=1) > 0)[0]
    Yinds = np.where(np.any(bwexp > 0, axis=0) > 0)[0]
    bw = bw[Xinds[0]:Xinds[-1] + 1, Yinds[0]:Yinds[-1] + 1]
    notbw = 1 - dilation(bw, disk(censorSize))
    data = np.array(images[:, Xinds[0]:Xinds[-1] + 1, Yinds[0]:Yinds[-1] + 1])
    bw = (bw > 0)
    notbw = (notbw > 0)
    ref = np.median(data[:500, :, :], axis=0)
    bwexp[Xinds[0]:Xinds[-1] + 1, Yinds[0]:Yinds[-1] + 1] = True
    output['bwexp'] = bwexp

    # visualize ROI
    # fig = plt.figure()
    # plt.subplot(131);plt.imshow(ref);plt.axis('image');plt.xlabel('mean Intensity')
    # plt.subplot(132);plt.imshow(bw);plt.axis('image');plt.xlabel('initial ROI')
    # plt.subplot(133);plt.imshow(notbw);plt.axis('image');plt.xlabel('background')
    # fig.suptitle('ROI selection')
    # plt.show()

    output['meanIM'] = np.mean(data, axis=0)
    data = np.reshape(data, (data.shape[0], -1))
    data = data - np.mean(data, 0)
    data = data - np.mean(data, 0)

    # remove low frequency components
    data_hp = highpassVideo(data.T, 1 / tau_lp, sampleRate).T    
    
    
    """
    data_pred = np.empty_like(data_hp)
    if highPassRegression:
        data_pred[:] = highpassVideo(data, 1 / tau_pred, sampleRate)
    else:
        data_pred[:] = data_hp
    """
    
    # initial trace
    if weights_init is None:
        t = np.nanmean(data_hp[:, bw.ravel()], 1)
    else:
        t = -np.matmul(data_hp, weights_init[1:])    # weights are negative 
    t = t - np.mean(t)

    # remove any variance in trace that can be predicted from the background principal components
    Ub, Sb, Vb = svds(data_hp[:, notbw.ravel()], nPC_bg)
    
    if use_Ridge:
        alpha = nPC_bg * Ridge_bg_coef  # = np.single(np.linalg.norm(Ub, ord='fro') ** 2) * Ridge_bg_coef
        reg = Ridge(alpha=alpha, fit_intercept=False, solver='lsqr').fit(Ub, t)
    else:
        reg = LinearRegression(fit_intercept=False).fit(Ub, t)
        
    
    t = np.double(t - np.matmul(Ub, reg.coef_))
    

    # find out spikes of initial trace
    Xspikes, spikeTimes, guessData, output['rawROI']['falsePosRate'], output['rawROI']['detectionRate'], \
    output['rawROI']['templates'], low_spk, _ = denoiseSpikes(-t, windowLength, sampleRate, False, 100)

    Xspikes = -Xspikes
    output['rawROI']['X'] = t.copy()
    output['rawROI']['Xspikes'] = Xspikes.copy()
    output['rawROI']['spikeTimes'] = spikeTimes.copy()
    output['rawROI']['spatialFilter'] = bw.copy()
    output['rawROI']['X'] = output['rawROI']['X'] * np.mean(t[output['rawROI']['spikeTimes']]) / np.mean(
        output['rawROI']['X'][output['rawROI']['spikeTimes']])  # correct shrinkage
    output['num_spikes'] = [spikeTimes.shape[0]]
    templates = output['rawROI']['templates']
    selectSpikes = np.zeros(Xspikes.shape)
    selectSpikes[spikeTimes] = 1
    sgn = np.mean(Xspikes[selectSpikes > 0])
    noise = np.std(Xspikes[selectSpikes == 0])
    snr = sgn / noise

    # prebuild the regression matrix
    # generate a predictor for ridge regression
    pred = np.empty_like(data_hp)
    pred[:] = data_hp
    pred = np.hstack((np.ones((data_hp.shape[0], 1), dtype=np.single), np.reshape
    (movie.gaussian_blur_2D(np.reshape(pred,
                                       (data_hp.shape[0], ref.shape[0], ref.shape[1])),
                            kernel_size_x=7, kernel_size_y=7, kernel_std_x=1.5,
                            kernel_std_y=1.5, borderType=cv2.BORDER_REPLICATE), data_hp.shape)))

    # Cross-validation of regularized regression parameters
    lambdamax = np.single(np.linalg.norm(pred[:, 1:], ord='fro') ** 2)
    lambdas = lambdamax * np.logspace(-4, -2, 3)
    I0 = np.eye(pred.shape[1], dtype=np.single)
    I0[0, 0] = 0

    if doCrossVal:
        # need to add
        print('doing cross validation')
    else:
        s_max = 1
        l_max = 2
        lambd = lambdas[l_max]
        sigma = sigmas[s_max]
        lambda_ix = l_max

    selectPred = np.ones(data_hp.shape[0])
    if highPassRegression:
        selectPred[:np.int16(sampleRate / 2 + 1)] = 0
        selectPred[-1 - np.int16(sampleRate / 2):] = 0
    sigma = sigmas[s_max]

    """
    pred = np.empty_like(data_pred)
    pred[:] = data_pred
    pred = np.hstack((np.ones((data_pred.shape[0], 1), dtype=np.single), np.reshape
    (movie.gaussian_blur_2D(np.reshape(pred,
                                       (data_pred.shape[0], ref.shape[0], ref.shape[1])),
                            kernel_size_x=np.int(2 * np.ceil(2 * sigma) + 1),
                            kernel_size_y=np.int(2 * np.ceil(2 * sigma) + 1),
                            kernel_std_x=sigma, kernel_std_y=sigma,
                            borderType=cv2.BORDER_REPLICATE), data_pred.shape)))
    """
    recon = np.empty_like(data_hp)
    recon[:] = data_hp
    recon = np.hstack((np.ones((data_hp.shape[0], 1), dtype=np.single), np.reshape
    (movie.gaussian_blur_2D(np.reshape(recon,
                                       (data_hp.shape[0], ref.shape[0], ref.shape[1])),
                            kernel_size_x=np.int(2 * np.ceil(2 * sigma) + 1),
                            kernel_size_y=np.int(2 * np.ceil(2 * sigma) + 1),
                            kernel_std_x=sigma, kernel_std_y=sigma,
                            borderType=cv2.BORDER_REPLICATE), data_hp.shape)))

    # Identify spatial filters with regularized regression
    for j in range(3 if args['weight_update'] == 'NMF' else 1):
        if j==0:
            initialGuess = guessData[:]
        else:
            guessData = initialGuess[:]
        for iteration in range(nIter):
            doPlot = False
            if iteration == nIter - 1:
                doPlot = False

            # print('Identifying spatial filters')
            # print(iteration)
         
            gD = np.single(guessData[selectPred>0])
            if args['weight_update'] == 'NMF':
                C = np.array([gD, np.ones_like(gD)])  # constant baselines as 2nd component
                CCt = C.dot(C.T)
                CY = C.dot(recon[:, 1:])
                A = np.minimum(np.linalg.inv(CCt).dot(CY), 0)
                for _ in range(5):
                    for m in range(2):
                        A[m] += (CY[m] - CCt[m].dot(A)) / CCt[m, m]
                        if m == 0:
                            A[m] = np.minimum(A[m], 0)
                weights = np.concatenate([[0], A[0]])
            else:
                Ri = Ridge(alpha=lambdas[l_max], fit_intercept=True, solver='lsqr')
                Ri.fit(recon, gD)
                weights = Ri.coef_
                weights[0] = Ri.intercept_

            X = np.matmul(recon, weights)
            X = X - np.mean(X)

            # for i in np.where(np.array([np.corrcoef(X, Ub[:,i])[0,1] for i in range(nPC_bg)]) > .4)[0]:
            #     Ub[:, i] -= Ub[:, i].dot(X) / X.dot(X) * X

            spatialFilter = np.empty_like(weights)
            spatialFilter[:] = weights
            spatialFilter = movie.gaussian_blur_2D(np.reshape(spatialFilter[1:],
                                                              ref.shape, order='C')[np.newaxis, :, :],
                                                   kernel_size_x=np.int(2 * np.ceil(2 * sigma) + 1),
                                                   kernel_size_y=np.int(2 * np.ceil(2 * sigma) + 1),
                                                   kernel_std_x=sigma, kernel_std_y=sigma,
                                                   borderType=cv2.BORDER_REPLICATE)[0]

            if use_Ridge:
                #alpha = np.single(np.linalg.norm(Ub, ord='fro') ** 2) * 0.01
                b = Ridge(alpha=alpha, fit_intercept=False, solver='lsqr').fit(Ub, X).coef_
            else:
                b = LinearRegression(fit_intercept=False).fit(Ub, X).coef_
            
            """
            if doPlot:
                plt.figure()
                plt.plot(X)
                plt.plot(np.matmul(Ub, b))
                plt.title('Denoised trace vs background')
                plt.show()
            """
            
            X = X - np.matmul(Ub, b)
       
            # correct shrinkage
            X = np.double(X * np.mean(t[spikeTimes]) / np.mean(X[spikeTimes]))

            # generate the new trace and the new denoised trace
            Xspikes, spikeTimes, guessData, falsePosRate, detectionRate, templates, _, thresh = denoiseSpikes(-X,
                                                                                                      windowLength,
                                                                                                      sampleRate, doPlot=False)

            selectSpikes = np.zeros(Xspikes.shape)
            selectSpikes[spikeTimes] = 1
            sgn = np.mean(Xspikes[selectSpikes > 0])
            noise = np.std(Xspikes[selectSpikes == 0])
            snr = sgn / noise

            output['num_spikes'].append(spikeTimes.shape[0])

            # ensure that the maximum of the spatial filter is within the ROI
        matrix = np.matmul(np.transpose(pred[:, 1:]), -guessData)
        sigmax = np.sqrt(np.sum(np.multiply(pred[:, 1:], pred[:, 1:]), axis=0))
        sigmay = np.sqrt(np.dot(guessData, guessData))
        IMcorr = matrix / sigmax / sigmay
        maxCorrInROI = np.max(IMcorr[bw.ravel()])
        if np.any(IMcorr[notbw.ravel()] > maxCorrInROI):
            output['passedLocalityTest'] = False
            recon -= np.outer(recon.dot(weights), weights) / weights.dot(weights)
        else:
            output['passedLocalityTest'] = True
            break

    # compute SNR
    selectSpikes = np.zeros(Xspikes.shape)
    selectSpikes[spikeTimes] = 1
    sgn = np.mean(Xspikes[selectSpikes > 0])
    noise = np.std(Xspikes[selectSpikes == 0])
    snr = sgn / noise
    output['snr'] = snr

    del pred
    del recon
    data_lp = data - data_hp

    # output
    output['y'] = X
    output['yFilt'] = -Xspikes
    output['ROI'] = np.transpose(np.vstack((Xinds[[0, -1]], Yinds[[0, -1]])))
    output['ROIbw'] = bw
    output['recons_signal'] = guessData        
    output['spatialFilter'] = -spatialFilter
    output['falsePosRate'] = falsePosRate
    output['detectionRate'] = detectionRate
    output['templates'] = templates
    output['spikeTimes'] = spikeTimes
    output['thresh'] = thresh
    output['F0'] = np.nanmean(data_lp[:, bw.flatten()] + output['meanIM'][bw][np.newaxis, :], 1)
    output['dFF'] = X / output['F0']
    output['rawROI']['dFF'] = output['rawROI']['X'] / output['F0']
    output['bg_pc'] = Ub  # background components
    output['low_spk'] = low_spk
    output['weights'] = weights
    output['cellN'] = cellN

    return output


def denoiseSpikes(data, windowLength, sampleRate=400, doPlot=True, doClip=150):
    """ Function for finding spikes and the temporal filter given one dimensional signals.
        Use function whitenedMatchedFilter to denoise spikes. Function getThresh
        helps to find the best threshold given height of spikes.

    Args:
        data: 1-D array
            one dimensional signal

        windowLength: int
            length of window size for temporal filter

        sampleRate: int, default 400
            number of samples per second in the video

        doPlot: boolean, default:True
            if Ture, will plot trace of signals and spiketimes, peak triggered
            average, histogram of heights,

        doClip: int, default:150
            maximum number of spikes accepted

    Returns:
        datafilt: 1-D array
            signals after whitened matched filter

        spikeTimes: 1-D array
            record of time of spikes

        guessData: 1-D array
            recovery of original signals

        falsePosRate: float
            possibility of misclassify noise as real spikes

        detectionRate: float
            possibility of real spikes being detected

        templates: 1-D array
            temporal filter which is the peak triggered average

        low_spk: boolean
            true if number of spikes is smaller than 30
    """

    # highpass filter and threshold
    bb, aa = signal.butter(1, 1 / (sampleRate / 2), 'high')  # 1Hz filter
    dataHP = signal.filtfilt(bb, aa, data, padtype='odd', padlen=3 * (max(len(bb), len(aa)) - 1)).flatten()

    pks = dataHP[signal.find_peaks(dataHP, height=None)[0]]

    thresh, _, _, low_spk = getThresh(pks, doClip, 0.25)

    locs = signal.find_peaks(dataHP, height=thresh)[0]

    # peak-traiggered average
    window = np.int64(np.arange(-windowLength, windowLength + 1, 1))

    locs = locs[np.logical_and(locs > (-window[0]), locs < (len(data) - window[-1]))]
    PTD = data[(locs[:, np.newaxis] + window)]
    PTA = np.mean(PTD, 0)

    # matched filter
    datafilt = whitenedMatchedFilter(data, locs, window)     # Note in this step we use convolution(or template matching) to make spikes more prominent

    # spikes detected after filter
    pks2 = datafilt[signal.find_peaks(datafilt, height=None)[0]]

    thresh2, falsePosRate, detectionRate, _ = getThresh(pks2, doClip=0, pnorm=0.5)  # doClip=0 means no clipping
    spikeTimes = signal.find_peaks(datafilt, height=thresh2)[0]

    guessData = np.zeros(data.shape)
    guessData[spikeTimes] = 1
    guessData = np.convolve(guessData, PTA, 'same')

    # filtering shrinks the data;
    # rescale so that the mean value at the peaks is same as in the input
    datafilt = datafilt * np.mean(data[spikeTimes]) / np.mean(datafilt[spikeTimes])
    thresh2 = thresh2 * np.mean(data[spikeTimes]) / np.mean(datafilt[spikeTimes])

    # output templates
    templates = PTA

    # plot three graphs
    if doPlot:
        plt.figure()
        plt.subplot(211)
        plt.hist(pks, 500)
        plt.axvline(x=thresh, c='r')
        plt.title('raw data')
        plt.subplot(212)
        plt.hist(pks2, 500)
        plt.axvline(x=thresh2, c='r')
        plt.title('after matched filter')
        plt.tight_layout()
        plt.show()

        plt.figure()
        plt.plot(np.transpose(PTD), c=[0.5, 0.5, 0.5])
        plt.plot(PTA, c='black', linewidth=2)
        plt.title('Peak-triggered average')
        plt.show()

        plt.figure()
        plt.subplot(211)
        plt.plot(data)
        plt.plot(locs, np.max(datafilt) * 1.1 * np.ones(locs.shape), color='r', marker='o', fillstyle='none',
                 linestyle='none')
        plt.plot(spikeTimes, np.max(datafilt) * 1 * np.ones(spikeTimes.shape), color='g', marker='o', fillstyle='none',
                 linestyle='none')
        plt.subplot(212)
        plt.plot(datafilt)
        plt.plot(locs, np.max(datafilt) * 1.1 * np.ones(locs.shape), color='r', marker='o', fillstyle='none',
                 linestyle='none')
        plt.plot(spikeTimes, np.max(datafilt) * 1 * np.ones(spikeTimes.shape), color='g', marker='o', fillstyle='none',
                 linestyle='none')
        plt.show()

    return datafilt, spikeTimes, guessData, falsePosRate, detectionRate, templates, low_spk, thresh2


def getThresh(pks, doClip, pnorm=0.5):
    """ Function for deciding threshold given heights of all peaks.

    Args:
        pks: 1-D array
            height of all peaks

        doClip: int

        pnorm: float, between 0 and 1, default is 0.5
            a variable deciding the amount of spikes chosen

    Returns:
        thresh: float
            threshold for choosing spikes

        falsePosRate: float
            possibility of misclassify noise as real spikes

        detectionRate: float
            possibility of real spikes being detected

        low_spk: boolean
            true if number of spikes is smaller than 30
    """
    # find median of the kernel density estimation of peak heights
    spread = np.array([pks.min(), pks.max()])
    spread = spread + np.diff(spread) * np.array([-0.05, 0.05])
    low_spk = False
    pts = np.linspace(spread[0], spread[1], 2001)
    kde = stats.gaussian_kde(pks)
    f = kde(pts)    
    xi = pts
    center = np.where(xi > np.median(pks))[0][0]

    fmodel = np.concatenate([f[0:center + 1], np.flipud(f[0:center])])
    if len(fmodel) < len(f):
        fmodel = np.append(fmodel, np.ones(len(f) - len(fmodel)) * min(fmodel))
    else:
        fmodel = fmodel[0:len(f)]

    # adjust the model so it doesn't exceed the data:
    csf = np.cumsum(f) / np.sum(f)
    csmodel = np.cumsum(fmodel) / np.max([np.sum(f), np.sum(fmodel)])
    lastpt = np.where(np.logical_and(csf[0:-1] > csmodel[0:-1] + np.spacing(1), csf[1:] < csmodel[1:]))[0]
    if not lastpt.size:
        lastpt = center
    else:
        lastpt = lastpt[0]
    fmodel[0:lastpt + 1] = f[0:lastpt + 1]
    fmodel[lastpt:] = np.minimum(fmodel[lastpt:], f[lastpt:])

    # find threshold
    csf = np.cumsum(f)
    csmodel = np.cumsum(fmodel)
    csf2 = csf[-1] - csf
    csmodel2 = csmodel[-1] - csmodel
    obj = csf2 ** pnorm - csmodel2 ** pnorm
    maxind = np.argmax(obj)
    thresh = xi[maxind]

    if np.sum(pks > thresh) < 30:
        low_spk = True
        print(
            'Very few spikes were detected at the desired sensitivity/specificity tradeoff. Adjusting threshold to take 30 largest spikes')
        thresh = np.percentile(pks, 100 * (1 - 30 / len(pks)))
    elif ((np.sum(pks > thresh) > doClip) & (doClip > 0)):
        print('Selecting top', doClip, 'spikes for template')
        thresh = np.percentile(pks, 100 * (1 - doClip / len(pks)))

    ix = np.argmin(np.abs(xi - thresh))
    falsePosRate = csmodel2[ix] / csf2[ix]
    detectionRate = (csf2[ix] - csmodel2[ix]) / np.max(csf2 - csmodel2)
    return thresh, falsePosRate, detectionRate, low_spk


def whitenedMatchedFilter(data, locs, window):
    """
    Function for using whitened matched filter to the original signal for better
    SNR. Use welch method to approximate the spectral density of the signal.
    Rescale the signal in frequency domain. After scaling, convolve the signal with
    peak-triggered-average to make spikes more prominent.
    """
    N = np.ceil(np.log2(len(data)))
    censor = np.zeros(len(data))
    censor[locs] = 1
    censor = np.int16(np.convolve(censor.flatten(), np.ones([1, len(window)]).flatten(), 'same'))
    censor = (censor < 0.5)
    noise = data[censor]

    _, pxx = signal.welch(noise, fs=2 * np.pi, window=signal.get_window('hamming', 1000), nfft=2 ** N, detrend=False,
                          nperseg=1000)
    Nf2 = np.concatenate([pxx, np.flipud(pxx[1:-1])])
    scaling_vector = 1 / np.sqrt(Nf2)

    cc = np.pad(data.copy(),(0,np.int(2**N-len(data))),'constant')    
    dd = (cv2.dft(cc,flags=cv2.DFT_SCALE+cv2.DFT_COMPLEX_OUTPUT)[:,0,:]*scaling_vector[:,np.newaxis])[:,np.newaxis,:]
    dataScaled = cv2.idft(dd)[:,0,0]
    PTDscaled = dataScaled[(locs[:, np.newaxis] + window)]
    PTAscaled = np.mean(PTDscaled, 0)
    datafilt = np.convolve(dataScaled, np.flipud(PTAscaled), 'same')
    datafilt = datafilt[:len(data)]
    return datafilt


def highpassVideo(video, freq, sampleRate):
    """
    Function for passing signals with frequency higher than freq
    """
    normFreq = freq / (sampleRate / 2)
    b, a = signal.butter(3, normFreq, 'high')
    videoFilt = np.single(signal.filtfilt(b, a, video, padtype='odd', padlen=3 * (max(len(b), len(a)) - 1)))
    del video
    return videoFilt








